@@ -508,10 +508,11 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
   int acc_exec = 0, acc_emul = 0;
   int ret = ORC_TEST_OK;
   int bad = 0;
+  int misalignment;
 
   ORC_DEBUG ("got here");
 
-  if (!(flags & ORC_TEST_FLAGS_BACKUP)) {
+  {
     OrcTarget *target;
     unsigned int flags;
 
@@ -547,17 +548,23 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
   orc_executor_set_m (ex, m);
   ORC_DEBUG("size %d %d", ex->n, ex->params[ORC_VAR_A1]);
 
+  misalignment = 0;
   for(i=0;i<ORC_N_VARIABLES;i++){
     if (program->vars[i].name == NULL) continue;
 
     if (program->vars[i].vartype == ORC_VAR_TYPE_SRC) {
-      src[i-ORC_VAR_S1] = orc_array_new (n, m, program->vars[i].size);
+      src[i-ORC_VAR_S1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_random (src[i-ORC_VAR_S1], &rand_context);
+      misalignment++;
     } else if (program->vars[i].vartype == ORC_VAR_TYPE_DEST) {
-      dest_exec[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size);
+      dest_exec[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_pattern (dest_exec[i], ORC_OOB_VALUE);
-      dest_emul[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size);
+      dest_emul[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_pattern (dest_emul[i], ORC_OOB_VALUE);
+      misalignment++;
     } else if (program->vars[i].vartype == ORC_VAR_TYPE_PARAM) {
       orc_executor_set_param (ex, i, 2);
     }
@@ -575,7 +582,11 @@ orc_test_compare_output_full (OrcProgram *program, int flags)
     }
   }
   ORC_DEBUG ("running");
-  orc_executor_run (ex);
+  if (flags & ORC_TEST_FLAGS_BACKUP) {
+    orc_executor_run_backup (ex);
+  } else {
+    orc_executor_run (ex);
+  }
   ORC_DEBUG ("done running");
   for(i=0;i<ORC_N_VARIABLES;i++){
     if (program->vars[i].vartype == ORC_VAR_TYPE_ACCUMULATOR) {
@@ -713,21 +724,25 @@ orc_test_get_program_for_opcode (OrcStaticOpcode *opcode)
 {
   OrcProgram *p;
   char s[40];
+  int flags = 0;
 
+  p = orc_program_new ();
   if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    if (opcode->src_size[1] == 0) {
-      p = orc_program_new_as (opcode->dest_size[0], opcode->src_size[0]);
-    } else {
-      p = orc_program_new_ass (opcode->dest_size[0], opcode->src_size[0],
-          opcode->src_size[1]);
-    }
+    orc_program_add_accumulator (p, opcode->dest_size[0], "d1");
   } else {
-    if (opcode->src_size[1] == 0) {
-      p = orc_program_new_ds (opcode->dest_size[0], opcode->src_size[0]);
-    } else {
-      p = orc_program_new_dss (opcode->dest_size[0], opcode->src_size[0],
-          opcode->src_size[1]);
-    }
+    orc_program_add_destination (p, opcode->dest_size[0], "d1");
+  }
+  if (opcode->dest_size[1] != 0) {
+    orc_program_add_destination (p, opcode->dest_size[1], "d2");
+  }
+  orc_program_add_source (p, opcode->src_size[0], "s1");
+  if (opcode->src_size[1] != 0) {
+    orc_program_add_source (p, opcode->src_size[1], "s2");
+  }
+
+  if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
+      (opcode->flags & ORC_STATIC_OPCODE_FLOAT_DEST)) {
+    flags = ORC_TEST_FLAGS_FLOAT;
   }
 
   sprintf(s, "test_%s", opcode->name);
@@ -748,26 +763,45 @@ orc_test_get_program_for_opcode_const (OrcStaticOpcode *opcode)
 {
   OrcProgram *p;
   char s[40];
+  int args[4] = { -1, -1, -1, -1 };
+  int flags;
+  int n_args = 0;
 
   if (opcode->src_size[1] == 0) {
     return NULL;
   }
 
+  p = orc_program_new ();
   if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    p = orc_program_new_as (opcode->dest_size[0], opcode->src_size[0]);
+    args[n_args++] =
+      orc_program_add_accumulator (p, opcode->dest_size[0], "d1");
   } else {
-    p = orc_program_new_ds (opcode->dest_size[0], opcode->src_size[0]);
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[0], "d1");
+  }
+  if (opcode->dest_size[1] != 0) {
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[1], "d2");
+  }
+  args[n_args++] =
+    orc_program_add_source (p, opcode->src_size[0], "s1");
+  args[n_args++] =
+    orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
+  if (opcode->src_size[2]) {
+    args[n_args++] =
+      orc_program_add_constant (p, opcode->src_size[2], 1, "c2");
+  }
+
+  if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
+      (opcode->flags & ORC_STATIC_OPCODE_FLOAT_DEST)) {
+    flags = ORC_TEST_FLAGS_FLOAT;
   }
-  orc_program_add_constant (p, opcode->src_size[1], 1, "c1");
 
   sprintf(s, "test_const_%s", opcode->name);
   orc_program_set_name (p, s);
 
-  if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    orc_program_append_str (p, opcode->name, "a1", "s1", "c1");
-  } else {
-    orc_program_append_str (p, opcode->name, "d1", "s1", "c1");
-  }
+  orc_program_append_2 (p, opcode->name, 0, args[0], args[1],
+      args[2], args[3]);
 
   return p;
 }
@@ -777,33 +811,58 @@ orc_test_get_program_for_opcode_param (OrcStaticOpcode *opcode)
 {
   OrcProgram *p;
   char s[40];
+  int args[4] = { -1, -1, -1, -1 };
+  int flags;
+  int n_args = 0;
 
   if (opcode->src_size[1] == 0) {
     return NULL;
   }
 
+  p = orc_program_new ();
   if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    p = orc_program_new_as (opcode->dest_size[0], opcode->src_size[0]);
+    args[n_args++] =
+      orc_program_add_accumulator (p, opcode->dest_size[0], "d1");
   } else {
-    p = orc_program_new_ds (opcode->dest_size[0], opcode->src_size[0]);
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[0], "d1");
+  }
+  if (opcode->dest_size[1] != 0) {
+    args[n_args++] =
+      orc_program_add_destination (p, opcode->dest_size[1], "d2");
+  }
+  args[n_args++] =
+    orc_program_add_source (p, opcode->src_size[0], "s1");
+  args[n_args++] =
+    orc_program_add_parameter (p, opcode->src_size[1], "p1");
+  if (opcode->src_size[2]) {
+    args[n_args++] =
+      orc_program_add_parameter (p, opcode->src_size[2], "p2");
   }
-  orc_program_add_parameter (p, opcode->src_size[1], "p1");
 
-  sprintf(s, "test_const_%s", opcode->name);
+  if ((opcode->flags & ORC_STATIC_OPCODE_FLOAT_SRC) ||
+      (opcode->flags & ORC_STATIC_OPCODE_FLOAT_DEST)) {
+    flags = ORC_TEST_FLAGS_FLOAT;
+  }
+
+  sprintf(s, "test_p_%s", opcode->name);
   orc_program_set_name (p, s);
 
-  if (opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) {
-    orc_program_append_str (p, opcode->name, "a1", "s1", "p1");
-  } else {
-    orc_program_append_str (p, opcode->name, "d1", "s1", "p1");
-  }
+  orc_program_append_2 (p, opcode->name, 0, args[0], args[1],
+      args[2], args[3]);
 
   return p;
 }
 
-
 void
 orc_test_performance (OrcProgram *program, int flags)
+{
+  orc_test_performance_full (program, flags, NULL);
+}
+
+double
+orc_test_performance_full (OrcProgram *program, int flags,
+    const char *target_name)
 {
   OrcExecutor *ex;
   int n;
@@ -815,20 +874,22 @@ orc_test_performance (OrcProgram *program, int flags)
   OrcCompileResult result;
   OrcProfile prof;
   double ave, std;
+  OrcTarget *target;
+  int misalignment;
 
   ORC_DEBUG ("got here");
 
+  target = orc_target_get_by_name (target_name);
+
   if (!(flags & ORC_TEST_FLAGS_BACKUP)) {
-    OrcTarget *target;
     unsigned int flags;
 
-    target = orc_target_get_default ();
     flags = orc_target_get_default_flags (target);
 
     result = orc_program_compile_full (program, target, flags);
     if (!ORC_COMPILE_RESULT_IS_SUCCESSFUL(result)) {
-      printf("compile failed\n");
-      return;
+      //printf("compile failed\n");
+      return 0;
     }
   }
 
@@ -836,7 +897,7 @@ orc_test_performance (OrcProgram *program, int flags)
     n = program->constant_n;
   } else {
     //n = 64 + (orc_random(&rand_context)&0xf);
-    n = 10000;
+    n = 1000;
   }
 
   ex = orc_executor_new (program);
@@ -853,17 +914,23 @@ orc_test_performance (OrcProgram *program, int flags)
   orc_executor_set_m (ex, m);
   ORC_DEBUG("size %d %d", ex->n, ex->params[ORC_VAR_A1]);
 
+  misalignment = 0;
   for(i=0;i<ORC_N_VARIABLES;i++){
     if (program->vars[i].name == NULL) continue;
 
     if (program->vars[i].vartype == ORC_VAR_TYPE_SRC) {
-      src[i-ORC_VAR_S1] = orc_array_new (n, m, program->vars[i].size);
+      src[i-ORC_VAR_S1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_random (src[i-ORC_VAR_S1], &rand_context);
+      misalignment++;
     } else if (program->vars[i].vartype == ORC_VAR_TYPE_DEST) {
-      dest_exec[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size);
+      dest_exec[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_pattern (dest_exec[i], ORC_OOB_VALUE);
-      dest_emul[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size);
+      dest_emul[i-ORC_VAR_D1] = orc_array_new (n, m, program->vars[i].size,
+          misalignment);
       orc_array_set_pattern (dest_emul[i], ORC_OOB_VALUE);
+      misalignment++;
     } else if (program->vars[i].vartype == ORC_VAR_TYPE_PARAM) {
       orc_executor_set_param (ex, i, 2);
     }
@@ -891,7 +958,7 @@ orc_test_performance (OrcProgram *program, int flags)
   ORC_DEBUG ("done running");
 
   orc_profile_get_ave_std (&prof, &ave, &std);
-  printf("%g %g\n", ave/(n*m), std/(n*m));
+  //printf("%g %g\n", ave/(n*m), std/(n*m));
 
   for(i=0;i<4;i++){
     if (dest_exec[i]) orc_array_free (dest_exec[i]);
@@ -903,6 +970,6 @@ orc_test_performance (OrcProgram *program, int flags)
 
   orc_executor_free (ex);
 
-  return;
+  return ave/(n*m);
 }
 
