@@ -71,7 +71,7 @@ BINARY_SL(minsl, "ORC_MIN(%s, %s)")
 BINARY_UL(minul, "ORC_MIN((orc_uint32)%s, (orc_uint32)%s)")
 BINARY_SL(mulll, "(%s * %s) & 0xffffffff")
 BINARY_SL(mulhsl, "((orc_int64)%s * (orc_int64)%s) >> 32")
-BINARY_UL(mulhul, "((orc_uint64)%s * (orc_uint64)%s) >> 32")
+BINARY_UL(mulhul, "((orc_uint64)(orc_uint32)%s * (orc_uint64)(orc_uint32)%s) >> 32")
 BINARY_SL(orl, "%s | %s")
 BINARY_SL(shll, "%s << %s")
 BINARY_SL(shrsl, "%s >> %s")
@@ -84,19 +84,30 @@ BINARY_SL(xorl, "%s ^ %s")
 
 UNARY_BW(convsbw, "%s")
 UNARY_BW(convubw, "(orc_uint8)%s")
-UNARY_WL(convswl, "%s")
-UNARY_WL(convuwl, "(orc_uint16)%s")
 UNARY_WB(convwb, "%s")
+UNARY_WB(convhwb, "((orc_uint16)%s)>>8")
 UNARY_WB(convssswb, "ORC_CLAMP_SB(%s)")
 UNARY_WB(convsuswb, "ORC_CLAMP_UB(%s)")
 UNARY_WB(convusswb, "ORC_CLAMP_SB((orc_uint16)%s)")
 UNARY_WB(convuuswb, "ORC_CLAMP_UB((orc_uint16)%s)")
+
+UNARY_WL(convswl, "%s")
+UNARY_WL(convuwl, "(orc_uint16)%s")
 UNARY_LW(convlw, "%s")
+UNARY_WB(convhlw, "((orc_uint32)%s)>>16")
 UNARY_LW(convssslw, "ORC_CLAMP_SW(%s)")
 UNARY_LW(convsuslw, "ORC_CLAMP_UW(%s)")
 UNARY_LW(convusslw, "ORC_CLAMP_SW((orc_uint32)%s)")
 UNARY_LW(convuuslw, "ORC_CLAMP_UW((orc_uint32)%s)")
 
+UNARY_LQ(convslq, "%s")
+UNARY_LQ(convulq, "(orc_uint32)%s")
+UNARY_LW(convql, "%s")
+UNARY_LW(convsssql, "ORC_CLAMP_SL(%s)")
+UNARY_LW(convsusql, "ORC_CLAMP_UL(%s)")
+UNARY_LW(convussql, "ORC_CLAMP_SL((orc_uint64)%s)")
+UNARY_LW(convuusql, "ORC_CLAMP_UL((orc_uint64)%s)")
+
 BINARY_BW(mulsbw, "%s * %s")
 BINARY_BW(mulubw, "(orc_uint8)%s * (orc_uint8)%s")
 BINARY_WL(mulswl, "%s * %s")
@@ -111,16 +122,16 @@ UNARY_LW(select1lw, "((orc_uint32)%s >> 16)&0xffff")
 UNARY_UW(swapw, "ORC_SWAP_W(%s)")
 UNARY_UL(swapl, "ORC_SWAP_L(%s)")
 
-BINARY_F(addf, "%s + %s")
-BINARY_F(subf, "%s - %s")
-BINARY_F(mulf, "%s * %s")
-BINARY_F(divf, "%s / %s")
-UNARY_F(sqrtf, "sqrt(%s)")
-BINARY_F(maxf, "ORC_MAX(%s,%s)")
-BINARY_F(minf, "ORC_MIN(%s,%s)")
-BINARY_FL(cmpeqf, "(%s == %s) ? (~0) : 0")
-BINARY_FL(cmpltf, "(%s < %s) ? (~0) : 0")
-BINARY_FL(cmplef, "(%s <= %s) ? (~0) : 0")
+BINARY_F(addf, "ORC_DENORMAL(ORC_DENORMAL(%s) + ORC_DENORMAL(%s))")
+BINARY_F(subf, "ORC_DENORMAL(ORC_DENORMAL(%s) - ORC_DENORMAL(%s))")
+BINARY_F(mulf, "ORC_DENORMAL(ORC_DENORMAL(%s) * ORC_DENORMAL(%s))")
+BINARY_F(divf, "ORC_DENORMAL(ORC_DENORMAL(%s) / ORC_DENORMAL(%s))")
+UNARY_F(sqrtf, "sqrt(ORC_DENORMAL(%s))")
+BINARY_F(maxf, "ORC_MAXF(ORC_DENORMAL(%s),ORC_DENORMAL(%s))")
+BINARY_F(minf, "ORC_MINF(ORC_DENORMAL(%s),ORC_DENORMAL(%s))")
+BINARY_FL(cmpeqf, "(ORC_DENORMAL(%s) == ORC_DENORMAL(%s)) ? (~0) : 0")
+BINARY_FL(cmpltf, "(ORC_DENORMAL(%s) < ORC_DENORMAL(%s)) ? (~0) : 0")
+BINARY_FL(cmplef, "(ORC_DENORMAL(%s) <= ORC_DENORMAL(%s)) ? (~0) : 0")
 UNARY_FL(convfl, "(int)%s")
 UNARY_LF(convlf, "%s")
 
