@@ -30,9 +30,11 @@
 
 void orc_compiler_assign_rules (OrcCompiler *compiler);
 void orc_compiler_global_reg_alloc (OrcCompiler *compiler);
+void orc_compiler_rewrite_insns (OrcCompiler *compiler);
 void orc_compiler_rewrite_vars (OrcCompiler *compiler);
 void orc_compiler_rewrite_vars2 (OrcCompiler *compiler);
 int orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j);
+int orc_compiler_new_temporary (OrcCompiler *compiler, int size);
 void orc_compiler_check_sizes (OrcCompiler *compiler);
 
 static char **_orc_compiler_flag_list;
@@ -106,10 +108,11 @@ orc_compiler_allocate_register (OrcCompiler *compiler, int data_reg)
     }
   }
 
-  /* FIXME on !x86, this is an error */
-  ORC_COMPILER_ERROR (compiler, "register overflow for %s reg",
-      data_reg ? "vector" : "gp");
-  compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+  if (data_reg || !compiler->allow_gp_on_stack) {
+    ORC_COMPILER_ERROR (compiler, "register overflow for %s reg",
+        data_reg ? "vector" : "gp");
+    compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+  }
 
   return 0;
 }
@@ -209,10 +212,11 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     ORC_LOG("variables");
     for(i=0;i<ORC_N_VARIABLES;i++){
       if (program->vars[i].size > 0) {
-        ORC_LOG("%d: %s %d %d", i,
+        ORC_LOG("%d: %s size %d type %d alloc %d", i,
             program->vars[i].name,
             program->vars[i].size,
-            program->vars[i].vartype);
+            program->vars[i].vartype,
+            program->vars[i].alloc);
       }
     }
     ORC_LOG("instructions");
@@ -232,6 +236,8 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
 
   memcpy (compiler->vars, program->vars,
       ORC_N_VARIABLES * sizeof(OrcVariable));
+  memset (compiler->vars + ORC_N_VARIABLES, 0,
+      (ORC_N_COMPILER_VARIABLES - ORC_N_VARIABLES) * sizeof(OrcVariable));
   compiler->n_temp_vars = program->n_temp_vars;
   compiler->n_dup_vars = 0;
 
@@ -244,6 +250,9 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
   orc_compiler_check_sizes (compiler);
   if (compiler->error) goto error;
 
+  orc_compiler_rewrite_insns (compiler);
+  if (compiler->error) goto error;
+
   orc_compiler_assign_rules (compiler);
   if (compiler->error) goto error;
 
@@ -253,15 +262,62 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
   orc_compiler_global_reg_alloc (compiler);
 
   orc_compiler_rewrite_vars2 (compiler);
+#if 0
+  {
+    ORC_ERROR("variables");
+    for(i=0;i<ORC_N_VARIABLES;i++){
+      if (compiler->vars[i].size > 0) {
+        ORC_ERROR("%d: %s size %d type %d alloc %d [%d,%d]", i,
+            compiler->vars[i].name,
+            compiler->vars[i].size,
+            compiler->vars[i].vartype,
+            compiler->vars[i].alloc,
+            compiler->vars[i].first_use,
+            compiler->vars[i].last_use);
+      }
+    }
+    ORC_ERROR("instructions");
+    for(i=0;i<compiler->n_insns;i++){
+      ORC_ERROR("%d: %s %d %d %d %d", i,
+          compiler->insns[i].opcode->name,
+          compiler->insns[i].dest_args[0],
+          compiler->insns[i].dest_args[1],
+          compiler->insns[i].src_args[0],
+          compiler->insns[i].src_args[1]);
+    }
+  }
+#endif
+
   if (compiler->error) goto error;
 
   ORC_INFO("allocating code memory");
   orc_compiler_allocate_codemem (compiler);
-
-  ORC_INFO("compiling for target");
-  compiler->target->compile (compiler);
   if (compiler->error) goto error;
 
+  ORC_INFO("compiling for target \"%s\"", compiler->target->name);
+  compiler->target->compile (compiler);
+  if (compiler->error) {
+    compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+    goto error;
+  }
+
+  program->orccode = orc_code_new ();
+  program->orccode->exec = program->code_exec;
+  program->orccode->code = program->code;
+  program->orccode->code_size = compiler->codeptr - program->code;
+  program->orccode->is_2d = program->is_2d;
+  program->orccode->constant_n = program->constant_n;
+  program->orccode->constant_m = program->constant_m;
+
+  program->orccode->n_insns = compiler->n_insns;
+  program->orccode->insns = malloc(sizeof(OrcInstruction) * compiler->n_insns);
+  memcpy (program->orccode->insns, compiler->insns,
+      sizeof(OrcInstruction) * compiler->n_insns);
+
+  program->orccode->vars = malloc (sizeof(OrcVariable) * ORC_N_COMPILER_VARIABLES);
+  memcpy (program->orccode->vars, compiler->vars,
+      sizeof(OrcVariable) * ORC_N_COMPILER_VARIABLES);
+
   program->asm_code = compiler->asm_code;
   program->code_size = compiler->codeptr - program->code;
 
@@ -299,25 +355,34 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
   for(i=0;i<compiler->n_insns;i++) {
     OrcInstruction *insn = compiler->insns + i;
     OrcStaticOpcode *opcode = insn->opcode;
+    int multiplier = 1;
+
+    if (insn->flags & ORC_INSTRUCTION_FLAG_X2) {
+      multiplier = 2;
+    } else if (insn->flags & ORC_INSTRUCTION_FLAG_X4) {
+      multiplier = 4;
+    }
 
     for(j=0;j<ORC_STATIC_OPCODE_N_DEST;j++){
       if (opcode->dest_size[j] == 0) continue;
-      if (opcode->dest_size[j] != compiler->vars[insn->dest_args[j]].size) {
+      if (multiplier * opcode->dest_size[j] !=
+          compiler->vars[insn->dest_args[j]].size) {
         ORC_COMPILER_ERROR(compiler, "size mismatch, opcode %s dest[%d] is %d should be %d",
             opcode->name, j, compiler->vars[insn->dest_args[j]].size,
-            opcode->dest_size[j]);
+            multiplier * opcode->dest_size[j]);
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         return;
       }
     }
     for(j=0;j<ORC_STATIC_OPCODE_N_SRC;j++){
       if (opcode->src_size[j] == 0) continue;
-      if (opcode->src_size[j] != compiler->vars[insn->src_args[j]].size &&
+      if (multiplier * opcode->src_size[j] !=
+          compiler->vars[insn->src_args[j]].size &&
           compiler->vars[insn->src_args[j]].vartype != ORC_VAR_TYPE_PARAM &&
           compiler->vars[insn->src_args[j]].vartype != ORC_VAR_TYPE_CONST) {
         ORC_COMPILER_ERROR(compiler, "size mismatch, opcode %s src[%d] is %d should be %d",
             opcode->name, j, compiler->vars[insn->src_args[j]].size,
-            opcode->src_size[j]);
+            multiplier * opcode->src_size[j]);
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         return;
       }
@@ -328,9 +393,169 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
             opcode->name);
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         return;
-
       }
     }
+    if (opcode->flags & ORC_STATIC_OPCODE_SCALAR &&
+        opcode->src_size[1] == 0 &&
+        compiler->vars[insn->src_args[0]].vartype != ORC_VAR_TYPE_PARAM &&
+        compiler->vars[insn->src_args[0]].vartype != ORC_VAR_TYPE_CONST) {
+      ORC_COMPILER_ERROR(compiler, "opcode %s requires const or param source",
+          opcode->name);
+      compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
+      return;
+    }
+  }
+}
+
+static OrcStaticOpcode *
+get_load_opcode_for_size (int size)
+{
+  switch (size) {
+    case 1:
+      return orc_opcode_find_by_name ("loadb");
+    case 2:
+      return orc_opcode_find_by_name ("loadw");
+    case 4:
+      return orc_opcode_find_by_name ("loadl");
+    case 8:
+      return orc_opcode_find_by_name ("loadq");
+    default:
+      ORC_ASSERT(0);
+  }
+  return NULL;
+}
+
+static OrcStaticOpcode *
+get_loadp_opcode_for_size (int size)
+{
+  switch (size) {
+    case 1:
+      return orc_opcode_find_by_name ("loadpb");
+    case 2:
+      return orc_opcode_find_by_name ("loadpw");
+    case 4:
+      return orc_opcode_find_by_name ("loadpl");
+    case 8:
+      return orc_opcode_find_by_name ("loadpq");
+    default:
+      ORC_ASSERT(0);
+  }
+  return NULL;
+}
+
+static OrcStaticOpcode *
+get_store_opcode_for_size (int size)
+{
+  switch (size) {
+    case 1:
+      return orc_opcode_find_by_name ("storeb");
+    case 2:
+      return orc_opcode_find_by_name ("storew");
+    case 4:
+      return orc_opcode_find_by_name ("storel");
+    case 8:
+      return orc_opcode_find_by_name ("storeq");
+    default:
+      ORC_ASSERT(0);
+  }
+  return NULL;
+}
+
+void
+orc_compiler_rewrite_insns (OrcCompiler *compiler)
+{
+  int i;
+  int j;
+  OrcStaticOpcode *opcode;
+  OrcProgram *program = compiler->program;
+
+  compiler->n_insns = 0;
+  for(j=0;j<program->n_insns;j++){
+    OrcInstruction insn;
+    OrcInstruction *xinsn;
+
+    memcpy (&insn, program->insns + j, sizeof(OrcInstruction));
+    opcode = insn.opcode;
+
+    if (!(opcode->flags & ORC_STATIC_OPCODE_LOAD)) {
+      for(i=0;i<ORC_STATIC_OPCODE_N_SRC;i++){
+        OrcVariable *var;
+
+        if (opcode->src_size[i] == 0) continue;
+        if (i > 0 && (opcode->flags & ORC_STATIC_OPCODE_SCALAR)) continue;
+
+        var = compiler->vars + insn.src_args[i];
+        if (var->vartype == ORC_VAR_TYPE_SRC ||
+            var->vartype == ORC_VAR_TYPE_DEST) {
+          OrcInstruction *cinsn;
+          
+          cinsn = compiler->insns + compiler->n_insns;
+          compiler->n_insns++;
+
+          cinsn->flags = insn.flags;
+          cinsn->flags |= ORC_INSN_FLAG_ADDED;
+          cinsn->flags &= ~(ORC_INSTRUCTION_FLAG_X2|ORC_INSTRUCTION_FLAG_X4);
+          cinsn->opcode = get_load_opcode_for_size (var->size);
+          cinsn->dest_args[0] = orc_compiler_new_temporary (compiler,
+              var->size);
+          cinsn->src_args[0] = insn.src_args[i];
+          insn.src_args[i] = cinsn->dest_args[0];
+        } else if (var->vartype == ORC_VAR_TYPE_CONST ||
+            var->vartype == ORC_VAR_TYPE_PARAM) {
+          OrcInstruction *cinsn;
+          int multiplier;
+
+          cinsn = compiler->insns + compiler->n_insns;
+          compiler->n_insns++;
+
+          multiplier = 1;
+          if (insn.flags & ORC_INSTRUCTION_FLAG_X2) {
+            multiplier = 2;
+          }
+          if (insn.flags & ORC_INSTRUCTION_FLAG_X4) {
+            multiplier = 4;
+          }
+
+          cinsn->flags = insn.flags;
+          cinsn->flags |= ORC_INSN_FLAG_ADDED;
+          cinsn->opcode = get_loadp_opcode_for_size (opcode->src_size[i]);
+          cinsn->dest_args[0] = orc_compiler_new_temporary (compiler,
+              opcode->src_size[i] * multiplier);
+          cinsn->src_args[0] = insn.src_args[i];
+          insn.src_args[i] = cinsn->dest_args[0];
+
+        }
+      }
+    }
+
+    xinsn = compiler->insns + compiler->n_insns;
+    memcpy (xinsn, &insn, sizeof(OrcInstruction));
+    compiler->n_insns++;
+
+    if (!(opcode->flags & ORC_STATIC_OPCODE_STORE)) {
+      for(i=0;i<ORC_STATIC_OPCODE_N_DEST;i++){
+        OrcVariable *var;
+
+        if (opcode->dest_size[i] == 0) continue;
+
+        var = compiler->vars + insn.dest_args[i];
+        if (var->vartype == ORC_VAR_TYPE_DEST) {
+          OrcInstruction *cinsn;
+          
+          cinsn = compiler->insns + compiler->n_insns;
+          compiler->n_insns++;
+
+          cinsn->flags = xinsn->flags;
+          cinsn->flags |= ORC_INSN_FLAG_ADDED;
+          cinsn->flags &= ~(ORC_INSTRUCTION_FLAG_X2|ORC_INSTRUCTION_FLAG_X4);
+          cinsn->opcode = get_store_opcode_for_size (var->size);
+          cinsn->src_args[0] = orc_compiler_new_temporary (compiler, var->size);
+          cinsn->dest_args[0] = xinsn->dest_args[i];
+          xinsn->dest_args[i] = cinsn->src_args[0];
+        }
+      }
+    }
+
   }
 }
 
@@ -354,6 +579,55 @@ orc_compiler_assign_rules (OrcCompiler *compiler)
   }
 }
 
+int
+orc_compiler_get_temp_reg (OrcCompiler *compiler)
+{
+  int j;
+
+  for(j=0;j<ORC_N_REGS;j++){
+    compiler->alloc_regs[j] = 0;
+  }
+  for(j=0;j<ORC_N_COMPILER_VARIABLES;j++){
+    if (!compiler->vars[j].alloc) continue;
+
+    ORC_DEBUG("var %d: %d  %d %d", j, compiler->vars[j].alloc,
+        compiler->vars[j].first_use,
+        compiler->vars[j].last_use);
+
+    if (compiler->vars[j].first_use == -1) {
+      compiler->alloc_regs[compiler->vars[j].alloc] = 1;
+    } else if (compiler->vars[j].first_use <= compiler->insn_index &&
+        compiler->vars[j].last_use >= compiler->insn_index) {
+      compiler->alloc_regs[compiler->vars[j].alloc] = 1;
+    }
+  }
+  for(j=0;j<compiler->n_constants;j++){
+    if (compiler->constants[j].alloc_reg) {
+      compiler->alloc_regs[compiler->constants[j].alloc_reg] = 1;
+    }
+  }
+
+  ORC_DEBUG("at insn %d %s", compiler->insn_index,
+      compiler->insns[compiler->insn_index].opcode->name);
+  for(j=0;j<8;j++){
+    ORC_DEBUG("xmm%d: %d %d", j, compiler->valid_regs[ORC_VEC_REG_BASE + j],
+        compiler->alloc_regs[ORC_VEC_REG_BASE + j]);
+  }
+
+  for(j=compiler->min_temp_reg;j<ORC_VEC_REG_BASE+16;j++){
+    if (compiler->valid_regs[j] && !compiler->alloc_regs[j]) {
+      compiler->min_temp_reg = j+1;
+      if (compiler->max_used_temp_reg < j) compiler->max_used_temp_reg = j;
+      return j;
+    }
+  }
+
+  ORC_COMPILER_ERROR(compiler,"no temporary register available");
+  compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+
+  return 0;
+}
+
 void
 orc_compiler_rewrite_vars (OrcCompiler *compiler)
 {
@@ -364,6 +638,10 @@ orc_compiler_rewrite_vars (OrcCompiler *compiler)
   int var;
   int actual_var;
 
+  for(j=0;j<ORC_N_COMPILER_VARIABLES;j++){
+    if (compiler->vars[j].alloc) continue;
+    compiler->vars[j].last_use = -1;
+  }
   for(j=0;j<compiler->n_insns;j++){
     insn = compiler->insns + j;
     opcode = insn->opcode;
@@ -376,6 +654,12 @@ orc_compiler_rewrite_vars (OrcCompiler *compiler)
       if (compiler->vars[var].vartype == ORC_VAR_TYPE_DEST) {
         compiler->vars[var].load_dest = TRUE;
       }
+      if (compiler->vars[var].vartype == ORC_VAR_TYPE_SRC ||
+          compiler->vars[var].vartype == ORC_VAR_TYPE_DEST ||
+          compiler->vars[var].vartype == ORC_VAR_TYPE_CONST ||
+          compiler->vars[var].vartype == ORC_VAR_TYPE_PARAM) {
+        continue;
+      }
 
       actual_var = var;
       if (compiler->vars[var].replaced) {
@@ -399,6 +683,9 @@ orc_compiler_rewrite_vars (OrcCompiler *compiler)
 
       var = insn->dest_args[k];
 
+      if (compiler->vars[var].vartype == ORC_VAR_TYPE_DEST) {
+        continue;
+      }
       if (compiler->vars[var].vartype == ORC_VAR_TYPE_SRC) {
         ORC_COMPILER_ERROR(compiler,"using src var as dest");
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
@@ -459,20 +746,13 @@ orc_compiler_global_reg_alloc (OrcCompiler *compiler)
   int i;
   OrcVariable *var;
 
-
-  for(i=0;i<ORC_N_VARIABLES;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     var = compiler->vars + i;
     if (var->name == NULL) continue;
     switch (var->vartype) {
       case ORC_VAR_TYPE_CONST:
-        var->first_use = -1;
-        var->last_use = -1;
-        var->alloc = orc_compiler_allocate_register (compiler, TRUE);
         break;
       case ORC_VAR_TYPE_PARAM:
-        var->first_use = -1;
-        var->last_use = -1;
-        var->alloc = orc_compiler_allocate_register (compiler, TRUE);
         break;
       case ORC_VAR_TYPE_SRC:
         var->ptr_register = orc_compiler_allocate_register (compiler, FALSE);
@@ -501,6 +781,24 @@ orc_compiler_global_reg_alloc (OrcCompiler *compiler)
     if (compiler->error) break;
   }
 
+  for(i=0;i<compiler->n_insns;i++){
+    OrcInstruction *insn = compiler->insns + i;
+    OrcStaticOpcode *opcode = insn->opcode;
+
+    if (opcode->flags & ORC_STATIC_OPCODE_INVARIANT) {
+      var = compiler->vars + insn->dest_args[0];
+
+      var->first_use = -1;
+      var->last_use = -1;
+      var->alloc = orc_compiler_allocate_register (compiler, TRUE);
+      insn->flags |= ORC_INSN_FLAG_INVARIANT;
+    }
+
+    if (opcode->flags & ORC_STATIC_OPCODE_ITERATOR) {
+      compiler->has_iterator_opcode = TRUE;
+    }
+  }
+
   if (compiler->alloc_loop_counter && !compiler->error) {
     compiler->loop_counter = orc_compiler_allocate_register (compiler, FALSE);
     /* FIXME massive hack */
@@ -525,6 +823,8 @@ orc_compiler_rewrite_vars2 (OrcCompiler *compiler)
      *  - src1 must be last_use
      *  - only one dest
      */
+    if (compiler->insns[j].flags & ORC_INSN_FLAG_INVARIANT) continue;
+
     if (!(compiler->insns[j].opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR)
         && compiler->insns[j].opcode->dest_size[1] == 0) {
       int src1 = compiler->insns[j].src_args[0];
@@ -552,15 +852,16 @@ orc_compiler_rewrite_vars2 (OrcCompiler *compiler)
       }
     }
 
-    for(i=0;i<ORC_N_VARIABLES;i++){
+    for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
       if (compiler->vars[i].name == NULL) continue;
+      if (compiler->vars[i].last_use == -1) continue;
       if (compiler->vars[i].first_use == j) {
         if (compiler->vars[i].alloc) continue;
         k = orc_compiler_allocate_register (compiler, TRUE);
         compiler->vars[i].alloc = k;
       }
     }
-    for(i=0;i<ORC_N_VARIABLES;i++){
+    for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
       if (compiler->vars[i].name == NULL) continue;
       if (compiler->vars[i].last_use == j) {
         compiler->alloc_regs[compiler->vars[i].alloc]--;
@@ -584,6 +885,20 @@ orc_compiler_dup_temporary (OrcCompiler *compiler, int var, int j)
   return i;
 }
 
+int
+orc_compiler_new_temporary (OrcCompiler *compiler, int size)
+{
+  int i = ORC_VAR_T1 + compiler->n_temp_vars + compiler->n_dup_vars;
+
+  compiler->vars[i].vartype = ORC_VAR_TYPE_TEMP;
+  compiler->vars[i].size = size;
+  compiler->vars[i].name = malloc (10);
+  sprintf(compiler->vars[i].name, "tmp%d", i);
+  compiler->n_dup_vars++;
+
+  return i;
+}
+
 void
 orc_compiler_dump_asm (OrcCompiler *compiler)
 {
@@ -635,10 +950,21 @@ orc_compiler_load_constant (OrcCompiler *compiler, int reg, int size,
   compiler->target->load_constant (compiler, reg, size, value);
 }
 
+int
+orc_compiler_get_temp_constant (OrcCompiler *compiler, int size, int value)
+{
+  int tmp;
+
+  tmp = orc_compiler_get_temp_reg (compiler);
+  orc_compiler_load_constant (compiler, tmp, size, value);
+  return tmp;
+}
+
 int
 orc_compiler_get_constant (OrcCompiler *compiler, int size, int value)
 {
   int i;
+  int tmp;
 
   if (size < 4) {
     if (size < 2) {
@@ -666,7 +992,52 @@ orc_compiler_get_constant (OrcCompiler *compiler, int size, int value)
   if (compiler->constants[i].alloc_reg != 0) {;
     return compiler->constants[i].alloc_reg;
   }
-  orc_compiler_load_constant (compiler, compiler->tmpreg, size, value);
-  return compiler->tmpreg;
+  tmp = orc_compiler_get_temp_reg (compiler);
+  orc_compiler_load_constant (compiler, tmp, size, value);
+  return tmp;
+}
+
+int
+orc_compiler_get_constant_reg (OrcCompiler *compiler)
+{
+  int j;
+
+  for(j=0;j<ORC_N_REGS;j++){
+    compiler->alloc_regs[j] = 0;
+  }
+  for(j=0;j<ORC_N_COMPILER_VARIABLES;j++){
+    if (!compiler->vars[j].alloc) continue;
+
+    ORC_DEBUG("var %d: %d  %d %d", j, compiler->vars[j].alloc,
+        compiler->vars[j].first_use,
+        compiler->vars[j].last_use);
+
+    if (compiler->vars[j].first_use == -1) {
+      compiler->alloc_regs[compiler->vars[j].alloc] = 1;
+    } else if (compiler->vars[j].last_use != -1) {
+      compiler->alloc_regs[compiler->vars[j].alloc] = 1;
+    }
+  }
+  for(j=0;j<compiler->n_constants;j++){
+    if (compiler->constants[j].alloc_reg) {
+      compiler->alloc_regs[compiler->constants[j].alloc_reg] = 1;
+    }
+  }
+  for(j=ORC_VEC_REG_BASE;j<=compiler->max_used_temp_reg;j++) {
+    compiler->alloc_regs[j] = 1;
+  }
+
+  for(j=0;j<8;j++){
+    ORC_DEBUG("xmm%d: %d %d", j, compiler->valid_regs[ORC_VEC_REG_BASE + j],
+        compiler->alloc_regs[ORC_VEC_REG_BASE + j]);
+  }
+
+  for(j=compiler->max_used_temp_reg;j<ORC_VEC_REG_BASE+16;j++){
+    if (compiler->valid_regs[j] && !compiler->alloc_regs[j]) {
+      return j;
+    }
+  }
+
+  return 0;
 }
 
