@@ -13,6 +13,7 @@
 #include <orc/orcutils.h>
 #include <orc/orcdebug.h>
 
+#define MMX 1
 #define SIZE 65536
 
 #define ORC_MMX_ALIGNED_DEST_CUTOFF 64
@@ -23,11 +24,13 @@ void orc_compiler_mmx_init (OrcCompiler *compiler);
 unsigned int orc_compiler_mmx_get_default_flags (void);
 void orc_compiler_mmx_assemble (OrcCompiler *compiler);
 void orc_compiler_mmx_register_rules (OrcTarget *target);
+void orc_mmx_emit_invariants (OrcCompiler *compiler);
 
 
 void orc_compiler_rewrite_vars (OrcCompiler *compiler);
 void orc_compiler_dump (OrcCompiler *compiler);
 void mmx_load_constant (OrcCompiler *compiler, int reg, int size, int value);
+static const char * mmx_get_flag_name (int shift);
 
 static OrcTarget mmx_target = {
   "mmx",
@@ -43,18 +46,30 @@ static OrcTarget mmx_target = {
   { { 0 } },
   0,
   NULL,
-  mmx_load_constant
-
+  mmx_load_constant,
+  mmx_get_flag_name
 };
 
 
+
 void
 orc_mmx_init (void)
 {
+#if defined(HAVE_AMD64) || defined(HAVE_I386)
+  /* initializes cache information */
+  orc_mmx_get_cpu_flags ();
+#endif
+
 #if defined(HAVE_I386)
+#ifndef MMX
+  if (!(orc_mmx_get_cpu_flags () & ORC_TARGET_MMX_MMXEXT)) {
+    mmx_target.executable = FALSE;
+  }
+#else
   if (!(orc_mmx_get_cpu_flags () & ORC_TARGET_MMX_MMX)) {
     mmx_target.executable = FALSE;
   }
+#endif
 #endif
 
   orc_target_register (&mmx_target);
@@ -70,18 +85,46 @@ orc_compiler_mmx_get_default_flags (void)
 #ifdef HAVE_AMD64
   flags |= ORC_TARGET_MMX_64BIT;
 #endif
-  flags &= ~ORC_TARGET_MMX_FRAME_POINTER;
+  if (_orc_compiler_flag_debug) {
+    flags |= ORC_TARGET_MMX_FRAME_POINTER;
+  }
   
 #if defined(HAVE_AMD64) || defined(HAVE_I386)
   flags |= orc_mmx_get_cpu_flags ();
+#else
+#ifndef MMX
+  flags |= ORC_TARGET_MMX_MMXEXT;
+  flags |= ORC_TARGET_MMX_SSE3;
+  flags |= ORC_TARGET_MMX_SSSE3;
 #else
   flags |= ORC_TARGET_MMX_MMX;
   flags |= ORC_TARGET_MMX_3DNOW;
+#endif
 #endif
 
   return flags;
 }
 
+static const char *
+mmx_get_flag_name (int shift)
+{
+  static const char *flags[] = {
+#ifndef MMX
+    "sse2", "sse3", "ssse3", "sse41", "sse42", "sse4a", "sse5",
+    "frame_pointer", "short_jumps", "64bit"
+#else
+    "mmx", "mmxext", "3dnow", "3dnowext", "ssse3", "sse41", "",
+    "frame_pointer", "short_jumps", "64bit"
+#endif
+  };
+
+  if (shift >= 0 && shift < sizeof(flags)/sizeof(flags[0])) {
+    return flags[shift];
+  }
+
+  return NULL;
+}
+
 void
 orc_compiler_mmx_init (OrcCompiler *compiler)
 {
@@ -104,9 +147,15 @@ orc_compiler_mmx_init (OrcCompiler *compiler)
     }
     compiler->valid_regs[X86_EDI] = 0;
     compiler->valid_regs[X86_ESP] = 0;
+#ifndef MMX
+    for(i=X86_MM0;i<X86_MM0+16;i++){
+      compiler->valid_regs[i] = 1;
+    }
+#else
     for(i=X86_MM0;i<X86_MM0+8;i++){
       compiler->valid_regs[i] = 1;
     }
+#endif
     compiler->save_regs[X86_EBX] = 1;
     compiler->save_regs[X86_EBP] = 1;
     compiler->save_regs[X86_R12] = 1;
@@ -133,9 +182,6 @@ orc_compiler_mmx_init (OrcCompiler *compiler)
     compiler->used_regs[i] = 0;
   }
 
-  compiler->tmpreg = X86_MM0;
-  compiler->valid_regs[compiler->tmpreg] = 0;
-
   compiler->gp_tmpreg = X86_ECX;
   compiler->valid_regs[compiler->gp_tmpreg] = 0;
 
@@ -152,26 +198,41 @@ orc_compiler_mmx_init (OrcCompiler *compiler)
 
   switch (orc_program_get_max_var_size (compiler->program)) {
     case 1:
-      compiler->loop_shift = 3;
+      compiler->loop_shift = 4;
       break;
     case 2:
-      compiler->loop_shift = 2;
+      compiler->loop_shift = 3;
       break;
     case 4:
-      compiler->loop_shift = 1;
+      compiler->loop_shift = 2;
       break;
     case 8:
-      compiler->loop_shift = 0;
+      compiler->loop_shift = 1;
       break;
     default:
       ORC_ERROR("unhandled max var size %d",
           orc_program_get_max_var_size (compiler->program));
       break;
   }
+#ifdef MMX
+  compiler->loop_shift--;
+#endif
 
-  compiler->unroll_shift = 1;
+  /* This limit is arbitrary, but some large functions run slightly
+     slower when unrolled (ginger Core2 6,15,6), and only some small
+     functions run faster when unrolled.  Most are the same speed. */
+  if (compiler->n_insns <= 10) {
+    compiler->unroll_shift = 1;
+  }
+  if (!compiler->long_jumps) {
+    compiler->unroll_shift = 0;
+  }
+  if (compiler->loop_shift == 0) {
+    /* FIXME something is broken with loop_shift=0, unroll_shift=1 */
+    compiler->unroll_shift = 0;
+  }
   compiler->alloc_loop_counter = TRUE;
-
+  compiler->allow_gp_on_stack = TRUE;
 }
 
 void
@@ -179,27 +240,47 @@ mmx_save_accumulators (OrcCompiler *compiler)
 {
   int i;
   int src;
+  int tmp;
 
-  for(i=0;i<ORC_N_VARIABLES;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     OrcVariable *var = compiler->vars + i;
 
     if (compiler->vars[i].name == NULL) continue;
     switch (compiler->vars[i].vartype) {
       case ORC_VAR_TYPE_ACCUMULATOR:
         src = compiler->vars[i].alloc;
+        tmp = orc_compiler_get_temp_reg (compiler);
 
-        orc_mmx_emit_pshufw (compiler, ORC_MMX_SHUF(3,2,3,2), src, compiler->tmpreg);
+#ifndef MMX
+        orc_mmx_emit_pshufd (compiler, ORC_MMX_SHUF(3,2,3,2), src, tmp);
+#else
+        orc_mmx_emit_pshufw (compiler, ORC_MMX_SHUF(3,2,3,2), src, tmp);
+#endif
 
         if (compiler->vars[i].size == 2) {
-          orc_mmx_emit_660f (compiler, "paddw", 0xfd, compiler->tmpreg, src);
+          orc_mmx_emit_660f (compiler, "paddw", 0xfd, tmp, src);
         } else {
-          orc_mmx_emit_660f (compiler, "paddd", 0xfe, compiler->tmpreg, src);
+          orc_mmx_emit_660f (compiler, "paddd", 0xfe, tmp, src);
         }
 
-        if (compiler->vars[i].size == 2) {
-          orc_mmx_emit_pshufw (compiler, ORC_MMX_SHUF(1,1,1,1), src, compiler->tmpreg);
+#ifndef MMX
+        orc_mmx_emit_pshufd (compiler, ORC_MMX_SHUF(1,1,1,1), src, tmp);
 
-          orc_mmx_emit_660f (compiler, "paddw", 0xfd, compiler->tmpreg, src);
+        if (compiler->vars[i].size == 2) {
+          orc_mmx_emit_660f (compiler, "paddw", 0xfd, tmp, src);
+        } else {
+          orc_mmx_emit_660f (compiler, "paddd", 0xfe, tmp, src);
+        }
+#endif
+
+        if (compiler->vars[i].size == 2) {
+#ifndef MMX
+          orc_mmx_emit_pshuflw (compiler, ORC_MMX_SHUF(1,1,1,1), src, tmp);
+#else
+          orc_mmx_emit_pshufw (compiler, ORC_MMX_SHUF(1,1,1,1), src, tmp);
+#endif
+
+          orc_mmx_emit_660f (compiler, "paddw", 0xfd, tmp, src);
         }
 
         if (compiler->vars[i].size == 2) {
@@ -225,57 +306,91 @@ mmx_save_accumulators (OrcCompiler *compiler)
 void
 mmx_load_constant (OrcCompiler *compiler, int reg, int size, int value)
 {
+  int i;
+
   if (size == 1) {
-    orc_mmx_emit_loadib (compiler, reg, value);
-  } else if (size == 2) {
-    orc_mmx_emit_loadiw (compiler, reg, value);
-  } else if (size == 4) {
-    orc_mmx_emit_loadil (compiler, reg, value);
-  } else {
-    ORC_COMPILER_ERROR(compiler, "unimplemented");
+    value &= 0xff;
+    value |= (value << 8);
+    value |= (value << 16);
+  }
+  if (size == 2) {
+    value &= 0xffff;
+    value |= (value << 16);
   }
 
+  ORC_ASM_CODE(compiler, "# loading constant %d 0x%08x\n", value, value);
+  if (value == 0) {
+    orc_mmx_emit_pxor(compiler, reg, reg);
+    return;
+  }
+  if (value == 0xffffffff) {
+    orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+    return;
+  }
+  if (compiler->target_flags & ORC_TARGET_MMX_SSSE3) {
+    if (value == 0x01010101) {
+      orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+      orc_mmx_emit_pabsb (compiler, reg, reg);
+      return;
+    }
+  }
+
+  for(i=1;i<32;i++){
+    orc_uint32 v;
+    v = (0xffffffff<<i);
+    if (value == v) {
+      orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+      orc_mmx_emit_pslld (compiler, i, reg);
+      return;
+    }
+    v = (0xffffffff>>i);
+    if (value == v) {
+      orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+      orc_mmx_emit_psrld (compiler, i, reg);
+      return;
+    }
+  }
+  for(i=1;i<16;i++){
+    orc_uint32 v;
+    v = (0xffff & (0xffff<<i)) | (0xffff0000 & (0xffff0000<<i));
+    if (value == v) {
+      orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+      orc_mmx_emit_psllw (compiler, i, reg);
+      return;
+    }
+    v = (0xffff & (0xffff>>i)) | (0xffff0000 & (0xffff0000>>i));
+    if (value == v) {
+      orc_mmx_emit_pcmpeqb (compiler, reg, reg);
+      orc_mmx_emit_psrlw (compiler, i, reg);
+      return;
+    }
+  }
+
+  orc_x86_emit_mov_imm_reg (compiler, 4, value, compiler->gp_tmpreg);
+  orc_x86_emit_mov_reg_mmx (compiler, compiler->gp_tmpreg, reg);
+#ifndef MMX
+  orc_mmx_emit_pshufd (compiler, ORC_MMX_SHUF(0,0,0,0), reg, reg);
+#else
+  orc_mmx_emit_pshufw (compiler, ORC_MMX_SHUF(1,0,1,0), reg, reg);
+#endif
 }
 
 void
 mmx_load_constants_outer (OrcCompiler *compiler)
 {
   int i;
-  for(i=0;i<ORC_N_VARIABLES;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     if (compiler->vars[i].name == NULL) continue;
     switch (compiler->vars[i].vartype) {
       case ORC_VAR_TYPE_CONST:
-        if (compiler->vars[i].size == 1) {
-          orc_mmx_emit_loadib (compiler, compiler->vars[i].alloc,
-              (int)compiler->vars[i].value);
-        } else if (compiler->vars[i].size == 2) {
-          orc_mmx_emit_loadiw (compiler, compiler->vars[i].alloc,
-              (int)compiler->vars[i].value);
-        } else if (compiler->vars[i].size == 4) {
-          orc_mmx_emit_loadil (compiler, compiler->vars[i].alloc,
-              (int)compiler->vars[i].value);
-        } else {
-          ORC_COMPILER_ERROR(compiler, "unimplemented");
-        }
         break;
       case ORC_VAR_TYPE_PARAM:
-        if (compiler->vars[i].size == 1) {
-          orc_mmx_emit_loadpb (compiler, compiler->vars[i].alloc, i);
-        } else if (compiler->vars[i].size == 2) {
-          orc_mmx_emit_loadpw (compiler, compiler->vars[i].alloc, i);
-        } else if (compiler->vars[i].size == 4) {
-          orc_mmx_emit_loadpl (compiler, compiler->vars[i].alloc, i);
-        } else if (compiler->vars[i].size == 8) {
-          orc_mmx_emit_loadpq (compiler, compiler->vars[i].alloc, i);
-        } else {
-          ORC_COMPILER_ERROR(compiler, "unimplemented");
-        }
         break;
       case ORC_VAR_TYPE_SRC:
       case ORC_VAR_TYPE_DEST:
         break;
       case ORC_VAR_TYPE_ACCUMULATOR:
-        orc_mmx_emit_660f (compiler, "pxor", 0xef,
+        orc_mmx_emit_pxor (compiler,
             compiler->vars[i].alloc, compiler->vars[i].alloc);
         break;
       case ORC_VAR_TYPE_TEMP:
@@ -285,13 +400,28 @@ mmx_load_constants_outer (OrcCompiler *compiler)
         break;
     }
   }
+
+  orc_mmx_emit_invariants (compiler);
+
+  /* FIXME move to a better place */
+  for(i=0;i<compiler->n_constants;i++){
+    compiler->constants[i].alloc_reg =
+      orc_compiler_get_constant_reg (compiler);
+  }
+
+  for(i=0;i<compiler->n_constants;i++){
+    if (compiler->constants[i].alloc_reg) {
+      mmx_load_constant (compiler, compiler->constants[i].alloc_reg,
+          4, compiler->constants[i].value);
+    }
+  }
 }
 
 void
 mmx_load_constants_inner (OrcCompiler *compiler)
 {
   int i;
-  for(i=0;i<ORC_N_VARIABLES;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     if (compiler->vars[i].name == NULL) continue;
     switch (compiler->vars[i].vartype) {
       case ORC_VAR_TYPE_CONST:
@@ -304,8 +434,6 @@ mmx_load_constants_inner (OrcCompiler *compiler)
           orc_x86_emit_mov_memoffset_reg (compiler, compiler->is_64bit ? 8 : 4,
               (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]), compiler->exec_reg,
               compiler->vars[i].ptr_register);
-        } else {
-          ORC_COMPILER_ERROR(compiler,"unimplemented");
         }
         break;
       case ORC_VAR_TYPE_ACCUMULATOR:
@@ -324,7 +452,7 @@ mmx_add_strides (OrcCompiler *compiler)
 {
   int i;
 
-  for(i=0;i<ORC_N_VARIABLES;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     if (compiler->vars[i].name == NULL) continue;
     switch (compiler->vars[i].vartype) {
       case ORC_VAR_TYPE_CONST:
@@ -351,99 +479,26 @@ mmx_add_strides (OrcCompiler *compiler)
   }
 }
 
-void
-orc_mmx_emit_load_src (OrcCompiler *compiler, OrcVariable *var, int offset)
-{
-  int ptr_reg;
-  if (var->ptr_register == 0) {
-    int i;
-    i = var - compiler->vars;
-    orc_x86_emit_mov_memoffset_reg (compiler, compiler->is_64bit ? 8 : 4,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor, arrays[i]),
-        compiler->exec_reg, compiler->gp_tmpreg);
-    ptr_reg = compiler->gp_tmpreg;
-  } else {
-    ptr_reg = var->ptr_register;
-  }
-  switch (var->size << compiler->loop_shift) {
-    case 1:
-      orc_x86_emit_mov_memoffset_reg (compiler, 1, offset, ptr_reg, compiler->gp_tmpreg);
-      orc_x86_emit_mov_reg_mmx (compiler, compiler->gp_tmpreg, var->alloc);
-      break;
-    case 2:
-      orc_x86_emit_mov_memoffset_reg (compiler, 2, offset, ptr_reg, compiler->gp_tmpreg);
-      orc_x86_emit_mov_reg_mmx (compiler, compiler->gp_tmpreg, var->alloc);
-      break;
-    case 4:
-      orc_x86_emit_mov_memoffset_mmx (compiler, 4, offset, ptr_reg, var->alloc,
-          var->is_aligned);
-      break;
-    case 8:
-      orc_x86_emit_mov_memoffset_mmx (compiler, 8, offset, ptr_reg, var->alloc,
-          var->is_aligned);
-      break;
-    case 16:
-      orc_x86_emit_mov_memoffset_mmx (compiler, 16, offset, ptr_reg, var->alloc,
-          var->is_aligned);
-      break;
-    default:
-      ORC_COMPILER_ERROR(compiler,"bad load size %d",
-          var->size << compiler->loop_shift);
-      break;
-  }
-}
-
-void
-orc_mmx_emit_store_dest (OrcCompiler *compiler, OrcVariable *var, int offset)
-{
-  int ptr_reg;
-  if (var->ptr_register == 0) {
-    orc_x86_emit_mov_memoffset_reg (compiler, compiler->is_64bit ? 8 : 4,
-        var->ptr_offset, compiler->exec_reg, compiler->gp_tmpreg);
-    ptr_reg = compiler->gp_tmpreg;
-  } else {
-    ptr_reg = var->ptr_register;
-  }
-  switch (var->size << compiler->loop_shift) {
-    case 1:
-      /* FIXME we might be using ecx twice here */
-      if (ptr_reg == compiler->gp_tmpreg) {
-        ORC_COMPILER_ERROR(compiler,"unimplemented");
-      }
-      orc_x86_emit_mov_mmx_reg (compiler, var->alloc, compiler->gp_tmpreg);
-      orc_x86_emit_mov_reg_memoffset (compiler, 1, compiler->gp_tmpreg, offset, ptr_reg);
-      break;
-    case 2:
-      /* FIXME we might be using ecx twice here */
-      if (ptr_reg == compiler->gp_tmpreg) {
-        ORC_COMPILER_ERROR(compiler,"unimplemented");
-      }
-      orc_x86_emit_mov_mmx_reg (compiler, var->alloc, compiler->gp_tmpreg);
-      orc_x86_emit_mov_reg_memoffset (compiler, 2, compiler->gp_tmpreg, offset, ptr_reg);
-      break;
-    case 4:
-      orc_x86_emit_mov_mmx_memoffset (compiler, 4, var->alloc, offset, ptr_reg,
-          var->is_aligned, var->is_uncached);
-      break;
-    case 8:
-      orc_x86_emit_mov_mmx_memoffset (compiler, 8, var->alloc, offset, ptr_reg,
-          var->is_aligned, var->is_uncached);
-      break;
-    case 16:
-      orc_x86_emit_mov_mmx_memoffset (compiler, 16, var->alloc, offset, ptr_reg,
-          var->is_aligned, var->is_uncached);
-      break;
-    default:
-      ORC_COMPILER_ERROR(compiler,"bad size");
-      break;
-  }
-}
-
 static int
 get_align_var (OrcCompiler *compiler)
 {
-  if (compiler->vars[ORC_VAR_D1].size) return ORC_VAR_D1;
-  if (compiler->vars[ORC_VAR_S1].size) return ORC_VAR_S1;
+  int i;
+  for(i=ORC_VAR_D1;i<=ORC_VAR_S8;i++){
+    if (compiler->vars[i].size == 0) continue;
+    if ((compiler->vars[i].size << compiler->loop_shift) >= 16) {
+      return i;
+    }
+  }
+  for(i=ORC_VAR_D1;i<=ORC_VAR_S8;i++){
+    if (compiler->vars[i].size == 0) continue;
+    if ((compiler->vars[i].size << compiler->loop_shift) >= 8) {
+      return i;
+    }
+  }
+  for(i=ORC_VAR_D1;i<=ORC_VAR_S8;i++){
+    if (compiler->vars[i].size == 0) continue;
+    return i;
+  }
 
   ORC_COMPILER_ERROR(compiler, "could not find alignment variable");
 
@@ -470,7 +525,7 @@ get_shift (int size)
 
 
 static void
-orc_emit_split_n_regions (OrcCompiler *compiler)
+orc_emit_split_3_regions (OrcCompiler *compiler)
 {
   int align_var;
   int align_shift;
@@ -536,6 +591,49 @@ orc_emit_split_n_regions (OrcCompiler *compiler)
   orc_x86_emit_label (compiler, 7);
 }
 
+static void
+orc_emit_split_2_regions (OrcCompiler *compiler)
+{
+  int align_var;
+  int align_shift;
+  int var_size_shift;
+
+  align_var = get_align_var (compiler);
+  var_size_shift = get_shift (compiler->vars[align_var].size);
+  align_shift = var_size_shift + compiler->loop_shift;
+
+  /* Calculate n2 */
+  orc_x86_emit_mov_memoffset_reg (compiler, 4,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg,
+      compiler->gp_tmpreg);
+  orc_x86_emit_mov_reg_reg (compiler, 4, compiler->gp_tmpreg, X86_EAX);
+  orc_x86_emit_sar_imm_reg (compiler, 4,
+      compiler->loop_shift + compiler->unroll_shift,
+      compiler->gp_tmpreg);
+  orc_x86_emit_mov_reg_memoffset (compiler, 4, compiler->gp_tmpreg,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2), compiler->exec_reg);
+
+  /* Calculate n3 */
+  orc_x86_emit_and_imm_reg (compiler, 4,
+      (1<<(compiler->loop_shift + compiler->unroll_shift))-1, X86_EAX);
+  orc_x86_emit_mov_reg_memoffset (compiler, 4, X86_EAX,
+      (int)ORC_STRUCT_OFFSET(OrcExecutor,counter3), compiler->exec_reg);
+}
+
+#ifndef MMX
+static int
+orc_program_has_float (OrcCompiler *compiler)
+{
+  int j;
+  for(j=0;j<compiler->n_insns;j++){
+    OrcInstruction *insn = compiler->insns + j;
+    OrcStaticOpcode *opcode = insn->opcode;
+    if (opcode->flags & ORC_STATIC_OPCODE_FLOAT) return TRUE;
+  }
+  return FALSE;
+}
+#endif
+
 #define LABEL_REGION1_SKIP 1
 #define LABEL_INNER_LOOP_START 2
 #define LABEL_REGION2_SKIP 3
@@ -548,6 +646,9 @@ orc_emit_split_n_regions (OrcCompiler *compiler)
 void
 orc_compiler_mmx_assemble (OrcCompiler *compiler)
 {
+#ifndef MMX
+  int set_mxcsr = FALSE;
+#endif
   int align_var;
 
   if (0 && orc_x86_assemble_copy_check (compiler)) {
@@ -560,8 +661,26 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
 
   compiler->vars[align_var].is_aligned = FALSE;
 
+  {
+    orc_mmx_emit_loop (compiler, 0, 0);
+
+    compiler->codeptr = compiler->program->code;
+    free (compiler->asm_code);
+    compiler->asm_code = NULL;
+    compiler->asm_code_len = 0;
+    memset (compiler->labels, 0, sizeof (compiler->labels));
+    compiler->n_fixups = 0;
+  }
+
   orc_x86_emit_prologue (compiler);
 
+#ifndef MMX
+  if (orc_program_has_float (compiler)) {
+    set_mxcsr = TRUE;
+    orc_mmx_set_mxcsr (compiler);
+  }
+#endif
+
   mmx_load_constants_outer (compiler);
 
   if (compiler->program->is_2d) {
@@ -589,12 +708,17 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
       compiler->program->constant_n <= ORC_MMX_ALIGNED_DEST_CUTOFF) {
     /* don't need to load n */
   } else if (compiler->loop_shift > 0) {
-    /* split n into three regions, with center region being aligned */
-    orc_emit_split_n_regions (compiler);
+    if (!compiler->has_iterator_opcode) {
+      /* split n into three regions, with center region being aligned */
+      orc_emit_split_3_regions (compiler);
+    } else {
+      orc_emit_split_2_regions (compiler);
+    }
   } else {
     /* loop shift is 0, no need to split */
     orc_x86_emit_mov_memoffset_reg (compiler, 4,
-        (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg, compiler->gp_tmpreg);
+        (int)ORC_STRUCT_OFFSET(OrcExecutor,n), compiler->exec_reg,
+        compiler->gp_tmpreg);
     orc_x86_emit_mov_reg_memoffset (compiler, 4, compiler->gp_tmpreg,
         (int)ORC_STRUCT_OFFSET(OrcExecutor,counter2), compiler->exec_reg);
   }
@@ -606,30 +730,42 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
     int n_left = compiler->program->constant_n;
     int save_loop_shift;
     int loop_shift;
-    int offset = 0;
+
+    compiler->offset = 0;
 
     save_loop_shift = compiler->loop_shift;
     while (n_left >= (1<<compiler->loop_shift)) {
       ORC_ASM_CODE(compiler, "# LOOP SHIFT %d\n", compiler->loop_shift);
-      orc_mmx_emit_loop (compiler, offset, 0);
+      orc_mmx_emit_loop (compiler, compiler->offset, 0);
 
       n_left -= 1<<compiler->loop_shift;
-      offset += 1<<compiler->loop_shift;
+      compiler->offset += 1<<compiler->loop_shift;
     }
     for(loop_shift = compiler->loop_shift-1; loop_shift>=0; loop_shift--) {
       if (n_left >= (1<<loop_shift)) {
         compiler->loop_shift = loop_shift;
         ORC_ASM_CODE(compiler, "# LOOP SHIFT %d\n", loop_shift);
-        orc_mmx_emit_loop (compiler, offset, 0);
+        orc_mmx_emit_loop (compiler, compiler->offset, 0);
         n_left -= 1<<loop_shift;
-        offset += 1<<loop_shift;
+        compiler->offset += 1<<loop_shift;
       }
     }
     compiler->loop_shift = save_loop_shift;
+
   } else {
     int ui, ui_max;
+    int emit_region1 = TRUE;
+    int emit_region3 = TRUE;
 
-    if (compiler->loop_shift > 0) {
+    if (compiler->has_iterator_opcode) {
+      emit_region1 = FALSE;
+    }
+    if (compiler->loop_shift == 0) {
+      emit_region1 = FALSE;
+      emit_region3 = FALSE;
+    }
+
+    if (emit_region1) {
       int save_loop_shift;
       int l;
 
@@ -668,9 +804,11 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
     orc_x86_emit_label (compiler, LABEL_INNER_LOOP_START);
     ui_max = 1<<compiler->unroll_shift;
     for(ui=0;ui<ui_max;ui++) {
-      orc_mmx_emit_loop (compiler, ui<<compiler->loop_shift,
+      compiler->offset = ui<<compiler->loop_shift;
+      orc_mmx_emit_loop (compiler, compiler->offset,
           (ui==ui_max-1) << (compiler->loop_shift + compiler->unroll_shift));
     }
+    compiler->offset = 0;
     if (compiler->loop_counter != ORC_REG_INVALID) {
       orc_x86_emit_add_imm_reg (compiler, 4, -1, compiler->loop_counter, TRUE);
     } else {
@@ -681,7 +819,7 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
     orc_x86_emit_jne (compiler, LABEL_INNER_LOOP_START);
     orc_x86_emit_label (compiler, LABEL_REGION2_SKIP);
 
-    if (compiler->loop_shift > 0) {
+    if (emit_region3) {
       int save_loop_shift;
       int l;
 
@@ -703,7 +841,7 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
     }
   }
 
-  if (compiler->program->is_2d) {
+  if (compiler->program->is_2d && compiler->program->constant_m != 1) {
     mmx_add_strides (compiler);
 
     orc_x86_emit_add_imm_memoffset (compiler, 4, -1,
@@ -715,7 +853,13 @@ orc_compiler_mmx_assemble (OrcCompiler *compiler)
 
   mmx_save_accumulators (compiler);
 
+#ifndef MMX
+  if (set_mxcsr) {
+    orc_mmx_restore_mxcsr (compiler);
+  }
+#else
   orc_x86_emit_emms (compiler);
+#endif
   orc_x86_emit_epilogue (compiler);
 
   orc_x86_do_fixups (compiler);
@@ -734,44 +878,17 @@ orc_mmx_emit_loop (OrcCompiler *compiler, int offset, int update)
     insn = compiler->insns + j;
     opcode = insn->opcode;
 
+    compiler->insn_index = j;
+
+    if (insn->flags & ORC_INSN_FLAG_INVARIANT) continue;
+
     ORC_ASM_CODE(compiler,"# %d: %s\n", j, insn->opcode->name);
 
-#if 0
-    /* set up args */
-    for(k=0;k<opcode->n_src + opcode->n_dest;k++){
-      args[k] = compiler->vars + insn->args[k];
-      ORC_ASM_CODE(compiler," %d", args[k]->alloc);
-      if (args[k]->is_chained) {
-        ORC_ASM_CODE(compiler," (chained)");
-      }
-    }
-    ORC_ASM_CODE(compiler,"\n");
-#endif
-
-    for(k=0;k<ORC_STATIC_OPCODE_N_SRC;k++){
-      OrcVariable *var = compiler->vars + insn->src_args[k];
-
-      if (opcode->src_size[k] == 0) continue;
-
-      switch (var->vartype) {
-        case ORC_VAR_TYPE_SRC:
-        case ORC_VAR_TYPE_DEST:
-          orc_mmx_emit_load_src (compiler, var, offset*var->size);
-          break;
-        case ORC_VAR_TYPE_CONST:
-          break;
-        case ORC_VAR_TYPE_PARAM:
-          break;
-        case ORC_VAR_TYPE_TEMP:
-          break;
-        default:
-          break;
-      }
-    }
+    compiler->min_temp_reg = ORC_VEC_REG_BASE;
 
     rule = insn->rule;
     if (rule && rule->emit) {
-      if (!(insn->opcode->flags & ORC_STATIC_OPCODE_ACCUMULATOR) &&
+      if (!(insn->opcode->flags & (ORC_STATIC_OPCODE_ACCUMULATOR|ORC_STATIC_OPCODE_LOAD|ORC_STATIC_OPCODE_STORE)) &&
           compiler->vars[insn->dest_args[0]].alloc !=
           compiler->vars[insn->src_args[0]].alloc) {
         orc_x86_emit_mov_mmx_reg_reg (compiler,
@@ -782,26 +899,10 @@ orc_mmx_emit_loop (OrcCompiler *compiler, int offset, int update)
     } else {
       ORC_COMPILER_ERROR(compiler,"No rule for: %s", opcode->name);
     }
-
-    for(k=0;k<ORC_STATIC_OPCODE_N_DEST;k++){
-      OrcVariable *var = compiler->vars + insn->dest_args[k];
-
-      if (opcode->dest_size[k] == 0) continue;
-
-      switch (var->vartype) {
-        case ORC_VAR_TYPE_DEST:
-          orc_mmx_emit_store_dest (compiler, var, offset*var->size);
-          break;
-        case ORC_VAR_TYPE_TEMP:
-          break;
-        default:
-          break;
-      }
-    }
   }
 
   if (update) {
-    for(k=0;k<ORC_N_VARIABLES;k++){
+    for(k=0;k<ORC_N_COMPILER_VARIABLES;k++){
       if (compiler->vars[k].name == NULL) continue;
       if (compiler->vars[k].vartype == ORC_VAR_TYPE_SRC ||
           compiler->vars[k].vartype == ORC_VAR_TYPE_DEST) {
@@ -820,3 +921,28 @@ orc_mmx_emit_loop (OrcCompiler *compiler, int offset, int update)
   }
 }
 
+void
+orc_mmx_emit_invariants (OrcCompiler *compiler)
+{
+  int j;
+  OrcInstruction *insn;
+  OrcStaticOpcode *opcode;
+  OrcRule *rule;
+
+  for(j=0;j<compiler->n_insns;j++){
+    insn = compiler->insns + j;
+    opcode = insn->opcode;
+
+    if (!(insn->flags & ORC_INSN_FLAG_INVARIANT)) continue;
+
+    ORC_ASM_CODE(compiler,"# %d: %s\n", j, insn->opcode->name);
+
+    rule = insn->rule;
+    if (rule && rule->emit) {
+      rule->emit (compiler, rule->emit_user, insn);
+    } else {
+      ORC_COMPILER_ERROR(compiler,"No rule for: %s", opcode->name);
+    }
+  }
+}
+
