@@ -18,8 +18,11 @@ typedef struct _OrcRuleSet OrcRuleSet;
 typedef struct _OrcConstant OrcConstant;
 typedef struct _OrcFixup OrcFixup;
 typedef struct _OrcTarget OrcTarget;
+typedef struct _OrcCode OrcCode;
 
 typedef void (*OrcOpcodeEmulateFunc)(OrcOpcodeExecutor *ex, void *user);
+typedef void (*OrcOpcodeEmulateNFunc)(OrcOpcodeExecutor *ex, int index, int n);
+typedef void (*OrcOpcodeEmulate16Func)(OrcOpcodeExecutor *ex);
 typedef void (*OrcRuleEmitFunc)(OrcCompiler *p, void *user, OrcInstruction *insn);
 typedef void (*OrcExecutorFunc)(OrcExecutor *ex);
 
@@ -28,9 +31,10 @@ typedef void (*OrcExecutorFunc)(OrcExecutor *ex);
 #define ORC_N_VARIABLES 64
 #define ORC_N_ARRAYS 12
 #define ORC_N_REGISTERS 20
-#define ORC_N_FIXUPS 40
+#define ORC_N_FIXUPS 100
 #define ORC_N_CONSTANTS 20
 #define ORC_N_LABELS 40
+#define ORC_N_COMPILER_VARIABLES (ORC_N_VARIABLES+32)
 
 #define ORC_GP_REG_BASE 32
 #define ORC_VEC_REG_BASE 64
@@ -43,6 +47,8 @@ typedef void (*OrcExecutorFunc)(OrcExecutor *ex);
 #define ORC_N_TARGETS 10
 #define ORC_N_RULE_SETS 10
 
+#define ORC_MAX_VAR_SIZE 8
+
 #define ORC_STRUCT_OFFSET(struct_type, member)    \
       ((long) ((unsigned int *) &((struct_type*) 0)->member))
 
@@ -75,6 +81,7 @@ enum {
   ORC_TARGET_C_C99 = (1<<0),
   ORC_TARGET_C_BARE = (1<<1),
   ORC_TARGET_C_NOEXEC = (1<<2),
+  ORC_TARGET_C_OPCODE = (1<<3),
   ORC_TARGET_FAST_NAN = (1<<30),
   ORC_TARGET_FAST_DENORMAL = (1<<31)
 };
@@ -85,7 +92,12 @@ enum {
 
 enum {
   ORC_TARGET_NEON_CLEAN_COMPILE = (1<<0),
-  ORC_TARGET_NEON_NEON = (1<<1)
+  ORC_TARGET_NEON_NEON = (1<<1),
+  ORC_TARGET_NEON_EDSP = (1<<2)
+};
+
+enum {
+  ORC_TARGET_ARM_EDSP = (1<<2)
 };
 
 typedef enum {
@@ -256,15 +268,18 @@ struct _OrcOpcodeSet {
 #define ORC_STATIC_OPCODE_FLOAT_DEST (1<<2)
 #define ORC_STATIC_OPCODE_FLOAT (ORC_STATIC_OPCODE_FLOAT_SRC|ORC_STATIC_OPCODE_FLOAT_DEST)
 #define ORC_STATIC_OPCODE_SCALAR (1<<3)
+#define ORC_STATIC_OPCODE_LOAD (1<<4)
+#define ORC_STATIC_OPCODE_STORE (1<<5)
+#define ORC_STATIC_OPCODE_INVARIANT (1<<6)
+#define ORC_STATIC_OPCODE_ITERATOR (1<<7)
 
 
 struct _OrcStaticOpcode {
   char name[16];
-  OrcOpcodeEmulateFunc emulate;
-  void *emulate_user;
   unsigned int flags;
   int dest_size[ORC_STATIC_OPCODE_N_DEST];
   int src_size[ORC_STATIC_OPCODE_N_SRC];
+  OrcOpcodeEmulateNFunc emulateN;
 };
 
 /**
@@ -279,8 +294,16 @@ struct _OrcInstruction {
   int src_args[ORC_STATIC_OPCODE_N_SRC];
 
   OrcRule *rule;
+  unsigned int flags;
 };
 
+#define ORC_INSTRUCTION_FLAG_X2 (1<<0)
+#define ORC_INSTRUCTION_FLAG_X4 (1<<1)
+
+#define ORC_INSN_FLAG_INVARIANT (1<<2)
+#define ORC_INSN_FLAG_ADDED (1<<3)
+
+
 /**
  * OrcConstant:
  *
@@ -314,7 +337,13 @@ struct _OrcFixup {
  */
 struct _OrcProgram {
   /*< private >*/
-  OrcInstruction insns[ORC_N_INSNS];
+  struct {
+    OrcStaticOpcode *opcode;
+    int dest_args[ORC_STATIC_OPCODE_N_DEST];
+    int src_args[ORC_STATIC_OPCODE_N_SRC];
+
+    OrcRule *rule;
+  } _unused[ORC_N_INSNS]; /* needed for ABI compatibility */
   int n_insns;
 
   OrcVariable vars[ORC_N_VARIABLES];
@@ -332,10 +361,17 @@ struct _OrcProgram {
   void *code_exec;
   int code_size;
 
+  OrcInstruction insns[ORC_N_INSNS];
+
   void *backup_func;
   int is_2d;
   int constant_n;
   int constant_m;
+
+  OrcCode *orccode;
+
+  /* Hide this here.  Belongs in a Parser object */
+  char *init_function;
 };
 
 /**
@@ -353,7 +389,7 @@ struct _OrcCompiler {
   OrcInstruction insns[ORC_N_INSNS];
   int n_insns;
 
-  OrcVariable vars[ORC_N_VARIABLES];
+  OrcVariable vars[ORC_N_COMPILER_VARIABLES];
   int n_temp_vars;
   int n_dup_vars;
 
@@ -388,12 +424,21 @@ struct _OrcCompiler {
   int gp_tmpreg;
 
   int insn_index;
+  int unroll_index;
   int need_mask_regs;
   int unroll_shift;
 
   int alloc_loop_counter;
+  int allow_gp_on_stack;
   int loop_counter;
   int size_region;
+  int has_iterator_opcode;
+
+  int offset;
+  int min_temp_reg;
+  int max_used_temp_reg;
+
+  int insn_shift; /* used when emitting rules */
 };
 
 #define ORC_SRC_ARG(p,i,n) ((p)->vars[(i)->src_args[(n)]].alloc)
@@ -412,6 +457,12 @@ struct _OrcOpcodeExecutor {
   /*< private >*/
   int src_values[ORC_STATIC_OPCODE_N_SRC];
   int dest_values[ORC_STATIC_OPCODE_N_DEST];
+
+  OrcOpcodeEmulateNFunc emulateN;
+
+  void *src_ptrs[ORC_STATIC_OPCODE_N_SRC];
+  void *dest_ptrs[ORC_STATIC_OPCODE_N_DEST];
+  int shift;
 };
 
 /**
@@ -465,6 +516,25 @@ struct _OrcExecutorAlt {
 #define ORC_EXECUTOR_M_INDEX(ex) ((ex)->params[ORC_VAR_A2])
 #define ORC_EXECUTOR_TIME(ex) ((ex)->params[ORC_VAR_A3])
 
+struct _OrcCode {
+  /*< private >*/
+  OrcCompileResult result;
+  char *name;
+
+  /* for execution */
+  OrcExecutorFunc exec;
+  unsigned char *code;
+  int code_size;
+
+  /* for emulation */
+  int n_insns;
+  OrcInstruction *insns;
+  OrcVariable *vars;
+  int is_2d;
+  int constant_n;
+  int constant_m;
+};
+
 /**
  * OrcTarget:
  *
@@ -506,8 +576,13 @@ void orc_program_set_constant_n (OrcProgram *program, int n);
 void orc_program_set_constant_m (OrcProgram *program, int m);
 
 void orc_program_append (OrcProgram *p, const char *opcode, int arg0, int arg1, int arg2);
+void orc_program_append_2 (OrcProgram *program, const char *name,
+    unsigned int flags, int arg0, int arg1, int arg2, int arg3);
 void orc_program_append_str (OrcProgram *p, const char *opcode,
     const char * arg0, const char * arg1, const char * arg2);
+void orc_program_append_str_2 (OrcProgram *program, const char *name,
+    unsigned int flags, const char *arg1, const char *arg2, const char *arg3,
+    const char *arg4);
 void orc_program_append_ds (OrcProgram *program, const char *opcode, int arg0,
     int arg1);
 void orc_program_append_ds_str (OrcProgram *p, const char *opcode,
@@ -558,6 +633,7 @@ void orc_executor_set_n (OrcExecutor *ex, int n);
 void orc_executor_set_m (OrcExecutor *ex, int m);
 void orc_executor_emulate (OrcExecutor *ex);
 void orc_executor_run (OrcExecutor *ex);
+void orc_executor_run_backup (OrcExecutor *ex);
 
 OrcOpcodeSet *orc_opcode_set_get (const char *name);
 OrcOpcodeSet *orc_opcode_set_get_nth (int opcode_major);
@@ -580,6 +656,10 @@ int orc_program_allocate_register (OrcProgram *program, int is_data);
 void orc_compiler_allocate_codemem (OrcCompiler *compiler);
 int orc_compiler_label_new (OrcCompiler *compiler);
 int orc_compiler_get_constant (OrcCompiler *compiler, int size, int value);
+int orc_compiler_get_temp_constant (OrcCompiler *compiler, int size, int value);
+int orc_compiler_get_temp_reg (OrcCompiler *compiler);
+int orc_compiler_get_constant_reg (OrcCompiler *compiler);
+
 
 const char *orc_program_get_asm_code (OrcProgram *program);
 const char *orc_target_get_asm_preamble (const char *target);
@@ -597,6 +677,10 @@ int orc_program_get_max_accumulator_size (OrcProgram *program);
 
 void orc_get_data_cache_sizes (int *level1, int *level2, int *level3);
 
+OrcCode * orc_code_new (void);
+void orc_code_free (OrcCode *code);
+
+
 #ifdef ORC_ENABLE_UNSTABLE_API
 
 int orc_compiler_flag_check (const char *flag);
