@@ -78,6 +78,87 @@ arm_rule_ ## opcode (OrcCompiler *p, void *user, OrcInstruction *insn) \
   orc_arm_emit(p,orc_arm_mull (0x00800090,cond,S,RdL,RdH,Rn,Rm)); \
 } while (0)
 
+static void
+arm_rule_loadpX (OrcCompiler *compiler, void *user, OrcInstruction *insn)
+{
+  if (compiler->vars[insn->src_args[0]].vartype == ORC_VAR_TYPE_CONST) {
+    orc_arm_emit_load_imm (compiler, compiler->vars[insn->dest_args[0]].alloc,
+        (int)compiler->vars[insn->src_args[0]].value);
+  } else {
+    orc_arm_loadw (compiler, compiler->vars[insn->dest_args[0]].alloc,
+        compiler->exec_reg,
+        (int)ORC_STRUCT_OFFSET(OrcExecutor, params[insn->src_args[0]]));
+  }
+}
+
+static void
+arm_rule_loadX (OrcCompiler *compiler, void *user, OrcInstruction *insn)
+{
+  int src = compiler->vars[insn->src_args[0]].ptr_register;
+  int dest = compiler->vars[insn->dest_args[0]].alloc;
+  int size;
+  orc_uint32 code;
+  int offset = 0;
+
+  size = (compiler->vars[insn->src_args[0]].size << compiler->loop_shift);
+
+  if (size == 4) {
+    code = 0xe5900000;
+    ORC_ASM_CODE(compiler,"  ldr %s, [%s, #%d]\n",
+        orc_arm_reg_name (dest),
+        orc_arm_reg_name (src), offset);
+  } else if (size == 2) {
+    code = 0xe1d000b0;
+    ORC_ASM_CODE(compiler,"  ldrh %s, [%s, #%d]\n",
+        orc_arm_reg_name (dest),
+        orc_arm_reg_name (src), offset);
+  } else {
+    code = 0xe5d00000;
+    ORC_ASM_CODE(compiler,"  ldrb %s, [%s, #%d]\n",
+        orc_arm_reg_name (dest),
+        orc_arm_reg_name (src), offset);
+  }
+  code |= (src&0xf) << 16;
+  code |= (dest&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+  orc_arm_emit (compiler, code);
+}
+
+static void
+arm_rule_storeX (OrcCompiler *compiler, void *user, OrcInstruction *insn)
+{
+  int src = compiler->vars[insn->src_args[0]].alloc;
+  int dest = compiler->vars[insn->dest_args[0]].ptr_register;
+  int size;
+  orc_uint32 code;
+  int offset = 0;
+
+  size = (compiler->vars[insn->src_args[0]].size << compiler->loop_shift);
+
+  if (size == 4) {
+    code = 0xe5800000;
+    ORC_ASM_CODE(compiler,"  str %s, [%s, #%d]\n",
+        orc_arm_reg_name (src),
+        orc_arm_reg_name (dest), offset);
+  } else if (size == 2) {
+    code = 0xe1c000b0;
+    ORC_ASM_CODE(compiler,"  strh %s, [%s, #%d]\n",
+        orc_arm_reg_name (src),
+        orc_arm_reg_name (dest), offset);
+  } else {
+    code = 0xe5c00000;
+    ORC_ASM_CODE(compiler,"  strb %s, [%s, #%d]\n",
+        orc_arm_reg_name (src),
+        orc_arm_reg_name (dest), offset);
+  }
+  code |= (dest&0xf) << 16;
+  code |= (src&0xf) << 12;
+  code |= (offset&0xf0) << 4;
+  code |= offset&0x0f;
+  orc_arm_emit (compiler, code);
+}
+
 void
 orc_arm_loadb (OrcCompiler *compiler, int dest, int src1, int offset)
 {
@@ -1499,61 +1580,32 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
 
   rule_set = orc_rule_set_new (orc_opcode_set_get("sys"), target, 0);
 
-  FAIL orc_rule_register (rule_set, "absb", arm_rule_absX, (void *)0);
-  orc_rule_register (rule_set, "addb", arm_rule_addb, NULL);
-  orc_rule_register (rule_set, "addssb", arm_rule_addssb, NULL);
-  orc_rule_register (rule_set, "addusb", arm_rule_addusb, NULL);
+  orc_rule_register (rule_set, "loadpb", arm_rule_loadpX, NULL);
+  orc_rule_register (rule_set, "loadpw", arm_rule_loadpX, NULL);
+  orc_rule_register (rule_set, "loadpl", arm_rule_loadpX, NULL);
+  orc_rule_register (rule_set, "loadb", arm_rule_loadX, NULL);
+  orc_rule_register (rule_set, "loadw", arm_rule_loadX, NULL);
+  orc_rule_register (rule_set, "loadl", arm_rule_loadX, NULL);
+  orc_rule_register (rule_set, "storeb", arm_rule_storeX, NULL);
+  orc_rule_register (rule_set, "storew", arm_rule_storeX, NULL);
+  orc_rule_register (rule_set, "storel", arm_rule_storeX, NULL);
+
   orc_rule_register (rule_set, "andb", arm_rule_andX, NULL);
   orc_rule_register (rule_set, "andnb", arm_rule_andnX, NULL);
   orc_rule_register (rule_set, "avgsb", arm_rule_avgX, (void *)3);
   orc_rule_register (rule_set, "avgub", arm_rule_avgX, (void *)0);
-  orc_rule_register (rule_set, "cmpeqb", arm_rule_cmpeqX, (void *)1);
-  orc_rule_register (rule_set, "cmpgtsb", arm_rule_cmpgtsX, (void *)1);
   orc_rule_register (rule_set, "copyb", arm_rule_copyX, NULL);
-  FAIL orc_rule_register (rule_set, "maxsb", arm_rule_maxsb, NULL);
-  FAIL orc_rule_register (rule_set, "maxub", arm_rule_maxub, NULL);
-  FAIL orc_rule_register (rule_set, "minsb", arm_rule_minsb, NULL);
-  FAIL orc_rule_register (rule_set, "minub", arm_rule_minub, NULL);
   orc_rule_register (rule_set, "orb", arm_rule_orX, NULL);
-  orc_rule_register (rule_set, "shlb", arm_rule_shlX, (void *)1);
-  FAIL orc_rule_register (rule_set, "shrsb", arm_rule_shrsX, (void *)1);
-  FAIL orc_rule_register (rule_set, "shrub", arm_rule_shruX, (void *)1);
-  FAIL orc_rule_register (rule_set, "signb", arm_rule_signX, (void *)0);
-  orc_rule_register (rule_set, "subb", arm_rule_subb, NULL);
-  orc_rule_register (rule_set, "subssb", arm_rule_subssb, NULL);
-  orc_rule_register (rule_set, "subusb", arm_rule_subusb, NULL);
   orc_rule_register (rule_set, "xorb", arm_rule_xorX, NULL);
-  orc_rule_register (rule_set, "mullb", arm_rule_mullb, NULL);
-  orc_rule_register (rule_set, "mulhsb", arm_rule_mulhsb, NULL);
-  FAIL orc_rule_register (rule_set, "mulhub", arm_rule_mulhub, NULL);
 
-  FAIL orc_rule_register (rule_set, "absw", arm_rule_absX, (void *)1);
-  orc_rule_register (rule_set, "addw", arm_rule_addw, NULL);
-  orc_rule_register (rule_set, "addssw", arm_rule_addssw, NULL);
-  orc_rule_register (rule_set, "addusw", arm_rule_addusw, NULL);
   orc_rule_register (rule_set, "andw", arm_rule_andX, NULL);
   orc_rule_register (rule_set, "andnw", arm_rule_andnX, NULL);
   FAIL orc_rule_register (rule_set, "avgsw", arm_rule_avgX, (void *)2);
   orc_rule_register (rule_set, "avguw", arm_rule_avgX, (void *)1);
-  orc_rule_register (rule_set, "cmpeqw", arm_rule_cmpeqX, (void *)2);
-  orc_rule_register (rule_set, "cmpgtsw", arm_rule_cmpgtsX, (void *)2);
   orc_rule_register (rule_set, "copyw", arm_rule_copyX, NULL);
-  FAIL orc_rule_register (rule_set, "maxsw", arm_rule_maxsw, NULL);
-  FAIL orc_rule_register (rule_set, "maxuw", arm_rule_maxuw, NULL);
-  FAIL orc_rule_register (rule_set, "minsw", arm_rule_minsw, NULL);
-  FAIL orc_rule_register (rule_set, "minuw", arm_rule_minuw, NULL);
   orc_rule_register (rule_set, "orw", arm_rule_orX, NULL);
-  orc_rule_register (rule_set, "shlw", arm_rule_shlX, (void *)2);
-  FAIL orc_rule_register (rule_set, "shrsw", arm_rule_shrsX, (void *)2);
-  orc_rule_register (rule_set, "shruw", arm_rule_shruX, (void *)2);
-  FAIL orc_rule_register (rule_set, "signw", arm_rule_signX, (void *)1);
-  orc_rule_register (rule_set, "subw", arm_rule_subw, NULL);
-  orc_rule_register (rule_set, "subssw", arm_rule_subssw, NULL);
-  orc_rule_register (rule_set, "subusw", arm_rule_subusw, NULL);
   orc_rule_register (rule_set, "xorw", arm_rule_xorX, NULL);
   orc_rule_register (rule_set, "mullw", arm_rule_mullw, NULL);
-  orc_rule_register (rule_set, "mulhsw", arm_rule_mulhsw, NULL);
-  FAIL orc_rule_register (rule_set, "mulhuw", arm_rule_mulhuw, NULL);
 
   FAIL orc_rule_register (rule_set, "absl", arm_rule_absl, NULL);
   orc_rule_register (rule_set, "addl", arm_rule_addl, NULL);
@@ -1563,8 +1615,6 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
   orc_rule_register (rule_set, "andnl", arm_rule_andnX, NULL);
   FAIL orc_rule_register (rule_set, "avgul", arm_rule_avgXl, NULL);
   FAIL orc_rule_register (rule_set, "avgsl", arm_rule_avgXl, NULL);
-  orc_rule_register (rule_set, "cmpeql", arm_rule_cmpeql, NULL);
-  orc_rule_register (rule_set, "cmpgtsl", arm_rule_cmpgtsl, NULL);
   orc_rule_register (rule_set, "copyl", arm_rule_copyX, NULL);
   orc_rule_register (rule_set, "maxsl", arm_rule_maxsl, NULL);
   orc_rule_register (rule_set, "maxul", arm_rule_maxul, NULL);
@@ -1574,35 +1624,83 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
   orc_rule_register (rule_set, "mulhsl", arm_rule_mulhsl, NULL);
   FAIL orc_rule_register (rule_set, "mulhul", arm_rule_mulhul, NULL);
   orc_rule_register (rule_set, "orl", arm_rule_orX, NULL);
-  orc_rule_register (rule_set, "shll", arm_rule_shlX, (void *)4);
-  orc_rule_register (rule_set, "shrsl", arm_rule_shrsX, (void *)4);
-  orc_rule_register (rule_set, "shrul", arm_rule_shruX, (void *)4);
   FAIL orc_rule_register (rule_set, "signl", arm_rule_signl, NULL);
   orc_rule_register (rule_set, "subl", arm_rule_subl, NULL);
   FAIL orc_rule_register (rule_set, "subssl", arm_rule_subssl, NULL);
   orc_rule_register (rule_set, "subusl", arm_rule_subusl, NULL);
   orc_rule_register (rule_set, "xorl", arm_rule_xorX, NULL);
 
-  orc_rule_register (rule_set, "convsbw", arm_rule_convsbw, NULL);
   FAIL orc_rule_register (rule_set, "convubw", arm_rule_convubw, NULL);
   FAIL orc_rule_register (rule_set, "convswl", arm_rule_convswl, NULL);
   FAIL orc_rule_register (rule_set, "convuwl", arm_rule_convuwl, NULL);
   orc_rule_register (rule_set, "convwb", arm_rule_convwb, NULL);
-  orc_rule_register (rule_set, "convssswb", arm_rule_convssswb, NULL);
+  orc_rule_register (rule_set, "convlw", arm_rule_convlw, NULL);
+
+  FAIL orc_rule_register (rule_set, "mulubw", arm_rule_mulubw, NULL);
+  FAIL orc_rule_register (rule_set, "mulswl", arm_rule_mulswl, NULL);
+  FAIL orc_rule_register (rule_set, "muluwl", arm_rule_muluwl, NULL);
+
+  rule_set = orc_rule_set_new (orc_opcode_set_get("sys"), target,
+      ORC_TARGET_ARM_EDSP);
+
+  FAIL orc_rule_register (rule_set, "absb", arm_rule_absX, (void *)0);
+  orc_rule_register (rule_set, "addb", arm_rule_addb, NULL);
+  orc_rule_register (rule_set, "addssb", arm_rule_addssb, NULL);
+  orc_rule_register (rule_set, "addusb", arm_rule_addusb, NULL);
+  orc_rule_register (rule_set, "cmpeqb", arm_rule_cmpeqX, (void *)1);
+  orc_rule_register (rule_set, "cmpgtsb", arm_rule_cmpgtsX, (void *)1);
+  FAIL orc_rule_register (rule_set, "maxsb", arm_rule_maxsb, NULL);
+  FAIL orc_rule_register (rule_set, "maxub", arm_rule_maxub, NULL);
+  FAIL orc_rule_register (rule_set, "minsb", arm_rule_minsb, NULL);
+  FAIL orc_rule_register (rule_set, "minub", arm_rule_minub, NULL);
+  orc_rule_register (rule_set, "mullb", arm_rule_mullb, NULL);
+  orc_rule_register (rule_set, "shlb", arm_rule_shlX, (void *)1);
+  FAIL orc_rule_register (rule_set, "shrsb", arm_rule_shrsX, (void *)1);
+  FAIL orc_rule_register (rule_set, "shrub", arm_rule_shruX, (void *)1);
+  FAIL orc_rule_register (rule_set, "signb", arm_rule_signX, (void *)0);
+  orc_rule_register (rule_set, "subb", arm_rule_subb, NULL);
+  orc_rule_register (rule_set, "subssb", arm_rule_subssb, NULL);
+  orc_rule_register (rule_set, "subusb", arm_rule_subusb, NULL);
+  orc_rule_register (rule_set, "mulhsb", arm_rule_mulhsb, NULL);
+  FAIL orc_rule_register (rule_set, "mulhub", arm_rule_mulhub, NULL);
+
+  FAIL orc_rule_register (rule_set, "absw", arm_rule_absX, (void *)1);
+  orc_rule_register (rule_set, "addssw", arm_rule_addssw, NULL);
+  orc_rule_register (rule_set, "addusw", arm_rule_addusw, NULL);
+  orc_rule_register (rule_set, "addw", arm_rule_addw, NULL);
+  orc_rule_register (rule_set, "cmpeqw", arm_rule_cmpeqX, (void *)2);
+  orc_rule_register (rule_set, "cmpgtsw", arm_rule_cmpgtsX, (void *)2);
+  FAIL orc_rule_register (rule_set, "maxsw", arm_rule_maxsw, NULL);
+  FAIL orc_rule_register (rule_set, "maxuw", arm_rule_maxuw, NULL);
+  FAIL orc_rule_register (rule_set, "minsw", arm_rule_minsw, NULL);
+  FAIL orc_rule_register (rule_set, "minuw", arm_rule_minuw, NULL);
+  orc_rule_register (rule_set, "mulsbw", arm_rule_mulsbw, NULL);
+  orc_rule_register (rule_set, "shlw", arm_rule_shlX, (void *)2);
+  FAIL orc_rule_register (rule_set, "shrsw", arm_rule_shrsX, (void *)2);
+  orc_rule_register (rule_set, "shruw", arm_rule_shruX, (void *)2);
+  FAIL orc_rule_register (rule_set, "signw", arm_rule_signX, (void *)1);
+  orc_rule_register (rule_set, "subssw", arm_rule_subssw, NULL);
+  orc_rule_register (rule_set, "subusw", arm_rule_subusw, NULL);
+  orc_rule_register (rule_set, "subw", arm_rule_subw, NULL);
+  orc_rule_register (rule_set, "mulhsw", arm_rule_mulhsw, NULL);
+  FAIL orc_rule_register (rule_set, "mulhuw", arm_rule_mulhuw, NULL);
+
+  orc_rule_register (rule_set, "cmpeql", arm_rule_cmpeql, NULL);
+  orc_rule_register (rule_set, "cmpgtsl", arm_rule_cmpgtsl, NULL);
+  orc_rule_register (rule_set, "shll", arm_rule_shlX, (void *)4);
+  orc_rule_register (rule_set, "shrsl", arm_rule_shrsX, (void *)4);
+  orc_rule_register (rule_set, "shrul", arm_rule_shruX, (void *)4);
+
+  orc_rule_register (rule_set, "convsbw", arm_rule_convsbw, NULL);
   orc_rule_register (rule_set, "convsuswb", arm_rule_convsuswb, NULL);
+  orc_rule_register (rule_set, "convssswb", arm_rule_convssswb, NULL);
   FAIL orc_rule_register (rule_set, "convusswb", arm_rule_convusswb, NULL);
   FAIL orc_rule_register (rule_set, "convuuswb", arm_rule_convuuswb, NULL);
-  orc_rule_register (rule_set, "convlw", arm_rule_convlw, NULL);
   FAIL orc_rule_register (rule_set, "convssslw", arm_rule_convssslw, NULL);
   orc_rule_register (rule_set, "convsuslw", arm_rule_convsuslw, NULL);
   FAIL orc_rule_register (rule_set, "convusslw", arm_rule_convusslw, NULL);
   FAIL orc_rule_register (rule_set, "convuuslw", arm_rule_convuuslw, NULL);
 
-  orc_rule_register (rule_set, "mulsbw", arm_rule_mulsbw, NULL);
-  FAIL orc_rule_register (rule_set, "mulubw", arm_rule_mulubw, NULL);
-  FAIL orc_rule_register (rule_set, "mulswl", arm_rule_mulswl, NULL);
-  FAIL orc_rule_register (rule_set, "muluwl", arm_rule_muluwl, NULL);
-
   orc_rule_register (rule_set, "mergewl", arm_rule_mergewl, NULL);
   FAIL orc_rule_register (rule_set, "mergebw", arm_rule_mergebw, NULL);
   orc_rule_register (rule_set, "select0wb", arm_rule_select0wb, NULL);
@@ -1612,5 +1710,6 @@ orc_compiler_orc_arm_register_rules (OrcTarget *target)
 
   orc_rule_register (rule_set, "swapw", arm_rule_swapw, NULL);
   orc_rule_register (rule_set, "swapl", arm_rule_swapl, NULL);
+
 }
 
