@@ -41,8 +41,9 @@ void orc_x86_emit_dec_memoffset (OrcCompiler *compiler, int size, int offset, in
 void orc_x86_emit_add_imm_memoffset (OrcCompiler *compiler, int size, int value, int offset, int reg);
 void orc_x86_emit_add_reg_memoffset (OrcCompiler *compiler, int size, int reg1, int offset, int reg);
 void orc_x86_emit_and_imm_memoffset (OrcCompiler *compiler, int size, int value, int offset, int reg);
-void orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg);
+void orc_x86_emit_add_imm_reg (OrcCompiler *compiler, int size, int value, int reg, orc_bool record);
 void orc_x86_emit_and_imm_reg (OrcCompiler *compiler, int size, int value, int reg);
+void orc_x86_emit_or_imm_reg (OrcCompiler *compiler, int size, int value, int reg);
 void orc_x86_emit_add_reg_reg (OrcCompiler *compiler, int size, int reg1, int reg2);
 void orc_x86_emit_sub_reg_reg (OrcCompiler *compiler, int size, int reg1, int reg2);
 void orc_x86_emit_imul_memoffset_reg (OrcCompiler *compiler, int size,
@@ -53,12 +54,15 @@ void orc_x86_emit_cmp_reg_memoffset (OrcCompiler *compiler, int size, int reg1,
     int offset, int reg);
 void orc_x86_emit_cmp_imm_memoffset (OrcCompiler *compiler, int size, int value,
     int offset, int reg);
+void orc_x86_emit_cmp_imm_reg (OrcCompiler *compiler, int size, int value, int reg);
 void orc_x86_emit_test_imm_memoffset (OrcCompiler *compiler, int size, int value,
     int offset, int reg);
 void orc_x86_emit_emms (OrcCompiler *compiler);
+void orc_x86_emit_rdtsc (OrcCompiler *compiler);
 void orc_x86_emit_rep_movs (OrcCompiler *compiler, int size);
 void orc_x86_emit_ret (OrcCompiler *compiler);
 void orc_x86_emit_jle (OrcCompiler *compiler, int label);
+void orc_x86_emit_jg (OrcCompiler *compiler, int label);
 void orc_x86_emit_je (OrcCompiler *compiler, int label);
 void orc_x86_emit_jne (OrcCompiler *compiler, int label);
 void orc_x86_emit_jmp (OrcCompiler *compiler, int label);
